#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-
#   Copyright (C) 2008-2010, 2013-2018, 2020-2024 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The command-line interface to the debugger.
"""
import os
import os.path as osp
import sys
import tempfile

import pyficache

from trepan.client import run
from trepan.clifns import whence_file
from trepan.debugger import Trepan
from trepan.exception import DebuggerQuit, DebuggerRestart
from trepan.interfaces.server import ServerInterface
from trepan.lib.file import is_compiled_py, readable
from trepan.misc import wrapped_lines
from trepan.options import postprocess_options, process_options
from trepan.version import __version__

package = "trepan"

# The name of the debugger we are currently going by.
__title__ = package


def main(dbg=None, sys_argv=list(sys.argv)):
    """Routine which gets run if we were invoked directly"""

    # Save the original just for use in the restart that works via exec.
    orig_sys_argv = list(sys_argv)
    opts, dbg_opts, sys_argv = process_options(__version__, sys_argv)

    if opts.server is not None:
        if opts.server == "tcp":
            connection_opts = {"IO": "TCP", "PORT": opts.port}
        else:
            connection_opts = {"IO": "FIFO"}
        intf = ServerInterface(connection_opts=connection_opts)
        dbg_opts["interface"] = intf
        if "FIFO" == intf.server_type:
            print(f"Starting FIFO server for process {os.getpid()}.")
        elif "TCP" == intf.server_type:
            print(f"Starting TCP server listening on port {intf.inout.PORT}.")
            pass
    elif opts.client:
        run(opts, sys_argv)
        return

    dbg_opts["orig_sys_argv"] = orig_sys_argv

    if dbg is None:
        dbg = Trepan(dbg_opts)
        dbg.core.add_ignore(main)

    postprocess_options(dbg, opts)

    # process_options has munged sys.argv to remove any options that
    # options that belong to this debugger. The original options to
    # invoke the debugger and script are in global sys_argv

    if len(sys_argv) == 0:
        # No program given to debug. Set to go into a command loop
        # anyway
        mainpyfile = None
    else:
        mainpyfile = sys_argv[0]  # Get script filename.
        if not osp.isfile(mainpyfile):
            mainpyfile = whence_file(mainpyfile)
            is_readable = readable(mainpyfile)
            if is_readable is None:
                print(f"{__title__}: Python script file '{mainpyfile}' does not exist")
                sys.exit(1)
            elif not is_readable:
                print(f"{__title__}: Can't read Python script file '{mainpyfile}'")
                sys.exit(1)
                return

        if is_compiled_py(mainpyfile):
            try:
                from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE, load_module
                from xdis.version_info import version_tuple_to_str

                (
                    python_version,
                    timestamp,
                    magic_int,
                    co,
                    is_pypy,
                    source_size,
                    sip_hash,
                ) = load_module(mainpyfile, code_objects=None, fast_load=False)
                if is_pypy != IS_PYPY:
                    print(f"Bytecode is pypy {is_pypy}, but we are {IS_PYPY}.")
                    print("For a cross-version debugger, use trepan-xpy with x-python.")
                    sys.exit(2)
                if python_version[:2] != PYTHON_VERSION_TRIPLE[:2]:
                    print(
                        "Bytecode is for version %s but we are version %s."
                        % (python_version, version_tuple_to_str())
                    )
                    print("For a cross-version debugger, use trepan-xpy with x-python.")
                    sys.exit(2)

                py_file = co.co_filename
                if osp.isabs(py_file):
                    try_file = py_file
                else:
                    mainpydir = osp.dirname(mainpyfile)
                    tag = sys.implementation.cache_tag
                    dirnames = (
                        [osp.join(mainpydir, tag), mainpydir]
                        + os.environ["PATH"].split(osp.pathsep)
                        + ["."]
                    )
                    try_file = whence_file(py_file, dirnames)

                if osp.isfile(try_file):
                    mainpyfile = try_file
                    pass
                else:
                    # Move onto the except branch
                    raise IOError(
                        f"Python file name embedded in code {try_file} not found"
                    )
            except IOError:
                decompiler = "uncompyle6"
                try:
                    if (3, 7) <= PYTHON_VERSION_TRIPLE <= (3, 8):
                        from uncompyle6 import decompile_file
                    else:
                        from decompyle3 import decompile_file

                        decompiler = "decompyle3"
                except ImportError:
                    print(
                        "%s: Compiled python file '%s', but %s not found"
                        % (__title__, mainpyfile, decompiler),
                        file=sys.stderr,
                    )
                    sys.exit(1)
                    return

                short_name = osp.basename(mainpyfile).strip(".pyc")
                fd = tempfile.NamedTemporaryFile(
                    suffix=".py",
                    prefix=short_name + "_",
                    delete=False,
                    dir=dbg.settings["tempdir"],
                )
                old_write = fd.file.write

                def write_wrapper(*args, **kwargs):
                    if isinstance(args[0], str):
                        new_args = list(args)
                        new_args[0] = args[0].encode("utf-8")
                        old_write(*new_args, **kwargs)
                    else:
                        old_write(*args, **kwargs)

                fd.file.write = write_wrapper

                # from io import StringIO
                # linemap_io = StringIO()
                try:
                    decompile_file(mainpyfile, fd.file, mapstream=fd)
                except Exception:
                    print(
                        f"{__title__}: error decompiling '{mainpyfile}'",
                        file=sys.stderr,
                    )
                    sys.exit(1)
                    return

                # # Get the line associations between the original and
                # # decompiled program
                # mapline = linemap_io.getvalue()
                # fd.write(mapline + "\n\n")
                # linemap = eval(mapline[3:])
                mainpyfile = fd.name
                fd.close()

                # Since we are actually running the recreated source,
                # there is little no need to remap line numbers.
                # The mapping is given at the end of the file.
                # However we should consider adding this information
                # and original file name.

                print(
                    "%s: couldn't find Python source so we recreated it at '%s'"
                    % (__title__, mainpyfile),
                    file=sys.stderr,
                )

                pass

        # If mainpyfile is an optimized Python script try to find and
        # use non-optimized alternative.
        mainpyfile_noopt = pyficache.resolve_name_to_path(mainpyfile)
        if mainpyfile != mainpyfile_noopt and readable(mainpyfile_noopt):
            print(f"{__title__}: Compiled Python script given and we can't use that.")
            print(f"{__title__}: Substituting non-compiled name: {mainpyfile_noopt}")
            mainpyfile = mainpyfile_noopt
            pass

        # Replace trepan's dir with script's dir in front of
        # module search path.
        sys.path[0] = dbg.main_dirname = osp.dirname(mainpyfile)

    # XXX If a signal has been received we continue in the loop, otherwise
    # the loop exits for some reason.
    dbg.sig_received = False

    # if not mainpyfile:
    #     print('For now, you need to specify a Python script name!')
    #     sys.exit(2)
    #     pass

    while True:
        # Run the debugged script over and over again until we get it
        # right.

        try:
            if dbg.program_sys_argv and mainpyfile:
                normal_termination = dbg.run_script(mainpyfile)
                if not normal_termination:
                    break
            else:
                dbg.core.execution_status = "No program"
                dbg.core.processor.process_commands()
                pass

            dbg.core.execution_status = "Terminated"
            dbg.intf[-1].msg("The program finished - quit or restart")
            dbg.core.processor.process_commands()
        except DebuggerQuit:
            break
        except DebuggerRestart:
            dbg.core.execution_status = "Restart requested"
            if dbg.program_sys_argv:
                sys.argv = list(dbg.program_sys_argv)
                part1 = f"Restarting {dbg.core.filename(mainpyfile)} with arguments:"
                args = " ".join(dbg.program_sys_argv[1:])
                dbg.intf[-1].msg(wrapped_lines(part1, args, dbg.settings["width"]))
            else:
                break
        except SystemExit:
            # In most cases SystemExit does not warrant a post-mortem session.
            break
        pass

    # Restore old sys.argv
    sys.argv = orig_sys_argv
    return


# When invoked as main program, invoke the debugger on a script
if __name__ == "__main__":
    main()



#!/usr/bin/env python
"General integration tests"
import unittest

import helper as Mhelper
from xdis import PYTHON_VERSION_TRIPLE


class GeneralTests(unittest.TestCase):
    def test_macro(self):
        """Test set/show highlight"""
        if PYTHON_VERSION_TRIPLE >= (3, 8):
            right_template = "%s-38.right"
        else:
            right_template = None
        result = Mhelper.run_debugger(
            testname="highlight",
            dbgr_opts="--basename " + "--highlight=plain --nx",
            python_file="gcd.py",
            right_template=right_template,
        )
        self.assertEqual(True, result, "'highlight' command comparision")
        return

    pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2023 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from pyficache import file2file_remap

from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowSubstitute(Mbase_subcmd.DebuggerSubcommand):
    """**show substitute** [*from-name1* *from_name2* ... ]
    If no *from-name* is given, then show all substitute commands

    Examples:
    ---------

        show substitute importlib._bootstrap /usr/lib/python3.10/importlib/_bootstrap.py
        show substitute
    """

    in_list = True
    min_abbrev = len("sub")

    short_help = "Set filename substitution"

    def run(self, args):
        if len(args) == 0:
            args = file2file_remap.keys()

        if len(args) == 0:
            self.msg("No file remappings in effect.")
        for from_path in args:
            self.msg(f"{from_path}:\t{file2file_remap.get(from_path, from_path)}")


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(ShowSubstitute, [])
    d = sub.proc.debugger
    sub.run(["show"])



"""Unit test for trepan.processor.command.finish"""

import inspect
from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.command.finish import FinishCommand


def test_finish():
    """Test processor.command.finish.FinishCommand.run()"""
    _, cp = setup_unit_test_debugger()

    command = FinishCommand(cp)
    for c in ((["finish", "5"], True), (["finish", "0*5+1"], True)):
        command.continue_running = False
        command.proc.stack = [(inspect.currentframe(), 14)]
        result = command.run(c[0])
        assert c[1] == result
        pass
    return



"""Unit test for trepan.lib.eval"""

from trepan.lib import eval as Meval


def test_extract_expression():
    for fragment, expect in (
        ("if condition(x):", "condition(x)"),
        ("elif is_magic(name):", "is_magic(name)"),
        ("while expression:", "expression"),
        ("for i in range(3):", "range(3)"),
        ("and x > 3", "x > 3"),
        ("or y < 3", "y < 3"),
        ("abc = 123", "123"),
        ("assert True", "True"),
        ("return return_value", "return_value"),
        ("nothing_to_be.done", "nothing_to_be.done"),
    ):
        assert expect == Meval.extract_expression(fragment)
        pass
    return



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2012-2013, 2015 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import columnize

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowAliases(Mbase_subcmd.DebuggerShowIntSubcommand):
    """**show aliases** [*alias* ...| *]

    Show command aliases. If parameters are given a list of all aliases and
    the command they run are printed. Alternatively one can list specific
    alias names for the commands those specific aliases are attached to.
    If instead of an alias `*` appears anywhere as an alias then just a list
    of aliases is printed, not what commands they are attached to.

    See also:
    ---------
    `alias`"""

    min_abbrev = len("al")
    short_help = "Show command aliases"
    run_cmd = False

    def _alias_header(self):
        self.section("%-10s : %s" % ("Alias", "Command"))
        self.msg("%-10s : %s" % ("-" * 10, "-" * 11))
        return

    def _alias_line(self, alias):
        self.msg("%-10s : %s" % (alias, self.proc.aliases[alias]))
        return

    def run(self, args):
        aliases = list(self.proc.aliases.keys())
        aliases.sort()
        if len(args) == 0:
            self._alias_header()
            for alias in aliases:
                self._alias_line(alias)
                pass
            return
        if "*" in args:
            self.section("Current aliases:")
            self.msg(columnize.columnize(aliases, lineprefix="    "))
        else:
            self._alias_header()
            for alias in args:
                if alias in aliases:
                    self._alias_line(alias)
                else:
                    self.errmsg("%s is not an alias" % alias)
                    pass
                pass
            return
        return


if __name__ == "__main__":
    from trepan.processor.command.show_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(ShowAliases)
    sub.run(["*"])
    sub.run(["s+", "n+"])
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2009, 2013, 2015, 2017, 2020-2021, 2023-2024
#   Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" A place for the debugger default settings """

# External Egg packages
import os

import tracer
from columnize import computed_displaywidth
from term_background import is_dark_background

width = computed_displaywidth()
is_dark_bg = is_dark_background()
default_style = "zenburn" if is_dark_bg else "tango"

# Below are the default debugger settings. The debugger object version
# of this may change. A setting is something a user may want to
# change, in contrast to settings that the debugger decides to set in
# the course of operation. For example, the maximum print width
# (width) is user settable, whereas whether the debugging program is
# running (execution_status), or traceback frame isn't user settable.
# Therefore, it doesn't appear below.  Some settings like the current frame
# (curframe) or the number of steps to skip before entering a command
# processor (step_ignore) are shared between the two.  They also don't
# generally appear as settings.

# fmt: off
DEBUGGER_SETTINGS = {
    # Emacs and old-style gdb annotate level. Used to annotate output
    # to make parsing inside Emacs easier and to allow Emacs to get
    # updated information (stack, local variables) without having to
    # poll for it.
    "annotate": 0,

    # Format style to use in showing disassembly
    "asmfmt": "extended",

    # Eval as Python the unrecognized debugger commands?
    "autoeval": True,

    # Run 'list' command every time we enter the debugger?
    "autolist": False,

    # Enter IPython every time we enter the debugger?
    # Note: only relevant if we have ipython installed. This takes
    # precedence over autopython.
    "autoipython": False,

    # Run 'info pc' command every time we enter the debugger?
    "autopc": False,

    # Enter Python every time we enter the debugger?
    "autopython": False,

    # Show basename only on filename output?
    # This option is useful in integration testing and
    # possibly to prepare example output for publication
    "basename": False,

    # Set echoing lines read from debugger?
    "cmdtrace": False,

    # confirm potentially dangerous operations?
    "confirm": True,

    # Debug macros?
    "debugmacro": False,

    # Debug the debugger?
    "dbg_trepan": False,

    # When True, consecutive stops must be on different
    # file/line positions.
    "different": True,

    # events is a set of events to process line-, call-, or return-like
    # tracing. See tracer.ALL_EVENT_NAMES and ALL_EVENTS
    # Note this is independent of printset which just prints the event.
    # This set controls entering the debugger command processor.
    "events": tracer.ALL_EVENTS,

    # Use terminal highlight? Acceptable values are
    #  'plain'   : no highlighting
    #  'dark'    : terminal highlighting for a dark background
    #  'light'   : terminal highlighting for a light background
    "highlight": "dark" if is_dark_bg else "light",

    # Where do we save the history?
    "histfile": None,

    # Save debugger history?
    "hist_save": True,

    # Show function calls/returns?
    "fntrace": False,

    # Number of lines to show by default in a 'list' command.
    "listsize": 10,

    # max length to show of parameter string
    "maxargstrsize": 100,

    # max length to in other strings
    "maxstring": 150,

    # printset is a set of events to print line-, call-, or return-like
    # tracing. See tracer.ALL_EVENT_NAMES and ALL_EVENTS. This only
    # has an effect if trace is set True.
    "printset": tracer.ALL_EVENTS,

    # If this is set True, debugger startup file, e.g. ".trepanrc" will
    # not be read/run.

    "nostartup": False,
    # Reread source file if we determine it has changed?

    "reload": False,
    # Skip instructions that make classes, functions, and closures?
    # (In the Python they are "class" and "def" statements)
    "skip": True,

    "step_ignore": 0,

    # Pygments style. Style is ignored if "highlight" setting
    # is "plain"
    "style": default_style,

    # Location to put temporary decompiled python files.
    # If value is None, use Python's defaults
    "tempdir": None,

    # print trace output?
    "trace": False,

    # The target maximum print length. Used for example in listing
    # arrays which are columnized.
    "width": width,
}

CLIENT_SOCKET_OPTS = {
    "HOST": "127.0.0.1",
    "PORT": 1027,
}  # Arbitrary non-privileged port


SERVER_SOCKET_OPTS = {
    "HOST": None,         # Symbolic name meaning all available interfaces
    "PORT": 1027,         # Arbitrary non-privileged port
    "reuse": "posix" == os.name,  # Allow port to be reused on close?
    "skew": +0,           # additional increment on socket tries
    "search_limit": 100,  # max number of ports to try
}
# fmt: on

# Default settings on the Debugger#start() method call
START_OPTS = {
    "add_hook_opts": tracer.DEFAULT_ADD_HOOK_OPTS,
    "backlevel": 0,  # trace caller and frames created from that
    "event_set": tracer.ALL_EVENTS,
    "force": False,  # Force a new event handler?
    "start": False,
}

# Default settings. on the Debugger#stop() method call.
STOP_OPTS = {"remove": False}

# Show it:
if __name__ == "__main__":
    import pprint

    for val in ["DEBUGGER_SETTINGS", "START_OPTS", "STOP_OPTS"]:
        print(f"{val}:")
        print(pprint.pformat(eval(val)))
        print("-" * 10)
        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowAutoPython(Mbase_subcmd.DebuggerShowBoolSubcommand):
    """**show autopython**

    Show whether we go into a python shell when automatically when the
    debugger is entered.

    Change with **set autopython**"""

    short_help = "Show automatic Python shell entry"
    min_abbrev = len("autopy")
    pass


if __name__ == "__main__":
    from trepan.processor.command.show_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(ShowAutoPython)
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2010, 2013-2015, 2018, 2023-2024
#   Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger class and top-level debugger functions.

This module contains the ``Trepan`` class and some top-level routines
for creating and invoking a debugger. Most of this module serves as:
  * a wrapper for `Debugger.core' routines,
  * a place to define debugger exceptions, and
  * debugger settings.

See also module ``cli`` which contains a command-line interface to debug
a Python script and `core' which contains the core debugging
start/stop and event-handling dispatcher and `client.py' which is a
user or client-side code for connecting to server'd debugged program.
"""

import sys
import types
from typing import Any, Callable

import pyficache
import tracefilter

# External Egg packages
import tracer

from trepan.exception import DebuggerQuit, DebuggerRestart
from trepan.interfaces.user import UserInterface
from trepan.lib.core import TrepanCore

# Default settings used here
from trepan.lib.default import DEBUGGER_SETTINGS, START_OPTS
from trepan.lib.sighandler import SignalManager
from trepan.misc import option_set

try:
    from readline import get_line_buffer
except ImportError:

    def get_line_buffer():
        return None

    pass

debugger_obj = None


class Trepan:
    """
    Class for a top-level object.
    """

    def __init__(self, opts=None):
        """Create a debugger object. But depending on the value of
        key 'start' inside hash 'opts', we may or may not initially
        start debugging.

        See also ``Debugger.start`` and ``Debugger.stop``.
        """

        self.mainpyfile = None
        self.thread = None
        self.eval_string = None
        self.settings = self.DEFAULT_INIT_OPTS["settings"].copy()

        def get_option(key: str) -> Any:
            return option_set(opts, key, self.DEFAULT_INIT_OPTS)

        def completer(text: str, state):
            return self.complete(text, state)

        # set the instance variables that come directly from options.
        for opt in ("settings", "orig_sys_argv", "from_ipython"):
            setattr(self, opt, get_option(opt))
            pass

        core_opts = {}
        for opt in (
            "ignore_filter",
            "proc_opts",
            "processor",
            "step_ignore",
            "processor",
        ):
            core_opts[opt] = get_option(opt)
            pass

        # How are I/O for this debugger handled? This should
        # be set before calling DebuggerCore.
        interface_opts = {
            "complete": completer,
            "debugger_name": "trepan3k",
        }
        # FIXME when I pass in opts=opts things break

        inp = opts.get("input", None) if opts else None
        interface = get_option("interface") or UserInterface(
            inp=inp, opts=interface_opts
        )
        self.intf = [interface]

        out = get_option("output")
        if out:
            self.intf[-1].output = out
            pass

        self.core = TrepanCore(self, core_opts)
        self.core.add_ignore(self.core.stop)

        # When set True, we'll also suspend our debug-hook tracing.
        # This gives us a way to prevent or allow self debugging.
        self.core.trace_hook_suspend = False

        if get_option("save_sys_argv"):
            # Save the debugged program's sys.argv? We do this so that
            # when the debugged script munges these, we have a good
            # copy to use for an exec restart
            self.program_sys_argv = list(sys.argv)
        else:
            self.program_sys_argv = None
            pass

        self.sigmgr = SignalManager(self)

        # Were we requested to activate immediately?
        if get_option("activate"):
            self.core.start(get_option("start_opts"))
            pass
        return

    def complete(self, last_token: str, state: int):
        """
        In place expansion of top-level debugger command
        for `last_token`` that we are in ``state``.
        """
        if hasattr(self.core.processor, "completer"):
            string_seen = get_line_buffer() or last_token
            results = self.core.processor.completer(string_seen, state)
            return results[state]
        return

    def run(self, cmd, start_opts=None, globals_=None, locals_=None):
        """Run debugger on string `cmd' using builtin function eval
        and if that builtin exec.  Arguments `globals_' and `locals_'
        are the dictionaries to use for local and global variables. By
        default, the value of globals is globals(), the current global
        variables. If `locals_' is not given, it becomes a copy of
        `globals_'.

        Debugger.core.start settings are passed via optional
        dictionary `start_opts'. Overall debugger settings are in
        ``Debugger.settings`` which changed after an instance is created
        . Also see `run_eval' if what you want to run is a
        run_eval'able expression have that result returned and
        `run_call' if you want to debug function run_call.
        """
        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(cmd, types.CodeType):
            self.eval_string = cmd
            cmd = cmd + "\n"
            pass
        retval = None
        self.core.start(start_opts)
        try:
            retval = eval(cmd, globals_, locals_)
        except SyntaxError:
            try:
                exec(cmd, globals_, locals_)
            except DebuggerQuit:
                pass
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return retval

    def run_exec(self, cmd, start_opts=None, globals_=None, locals_=None):
        """Run debugger on string `cmd' which will executed via the
        builtin function exec. Arguments `globals_' and `locals_' are
        the dictionaries to use for local and global variables. By
        default, the value of globals is globals(), the current global
        variables. If `locals_' is not given, it becomes a copy of
        `globals_'.

        ``Debugger.core.start`` settings are passed via optional
        dictionary `start_opts'. Overall debugger settings are in
        ``Debugger.settings`` which changed after an instance is created.
        See `run_eval' if what you want to run is an
         expression that should be ``eval``d and have that result returned.
        See ``run_call`` if you want to debug function ``run_call()``.
        """
        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(cmd, types.CodeType):
            cmd = cmd + "\n"
            pass
        self.core.start(start_opts)
        try:
            exec(cmd, globals_, locals_)
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return

    def run_call(self, func: Callable, *args, start_opts=None, **kwds):
        """Run debugger on function call: `func(*args, **kwds)'

        See also ``run_eval`` if what you want to run is an eval'able
        expression have that result returned and ``run``if you want to
        debug a statement via ``exec``.
        """
        res = None
        self.core.start(opts=start_opts)
        try:
            res = func(*args, **kwds)
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return res

    def run_eval(self, expr, start_opts=None, globals_=None, locals_=None):
        """Run debugger on string `expr' which will executed via the
        built-in Python function: eval; `globals_' and `locals_' are
        the dictionaries to use for local and global variables. If
        `globals' is not given, __main__.__dict__ (the current global
        variables) is used. If `locals_' is not given, it becomes a
        copy of `globals_'.

        See also `run_call' if what you to debug a function call and
        `run' if you want to debug general Python statements.
        """
        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(expr, types.CodeType):
            self.eval_string = expr
            expr = expr + "\n"
            pass
        retval = None
        self.core.start(start_opts)
        try:
            retval = eval(expr, globals_, locals_)
        except DebuggerQuit:
            pass
        finally:
            pyficache.remove_remap_file("<string>")
            self.core.stop()
        return retval

    def run_script(self, filename, start_opts=None, globals_=None, locals_=None):
        """Run debugger on Python script `filename'. The script may
        inspect sys.argv for command arguments. `globals_' and
        `locals_' are the dictionaries to use for local and global
        variables. If `globals' is not given, globals() (the current
        global variables) is used. If `locals_' is not given, it
        becomes a copy of `globals_'.

        True is returned if the program terminated normally and False
        if the debugger initiated a quit or the program did not normally
        terminate.

        See also `run_call' if what you to debug a function call,
        `run_eval' if you want to debug an expression, and `run' if you
        want to debug general Python statements not inside a file.
        """
        self.mainpyfile = self.core.canonic(filename)

        # Start with fresh empty copy of globals and locals and tell the script
        # that it's being run as __main__ to avoid scripts being able to access
        # the debugger namespace.
        if globals_ is None:
            import __main__  # NOQA

            globals_ = {
                "__name__": "__main__",
                "__file__": self.mainpyfile,
                "__builtins__": __builtins__,
            }  # NOQA
        if locals_ is None:
            locals_ = globals_
        retval = False
        self.core.execution_status = "Running"
        try:
            compiled = compile(open(self.mainpyfile).read(), self.mainpyfile, "exec")
        except SyntaxError:
            self.intf[0].errmsg(f"Python can't compile {self.mainpyfile}")
            self.intf[0].errmsg(sys.exc_info()[1])
            retval = False
            pass
        except UnicodeDecodeError:
            self.intf[0].errmsg(
                "File %s can't be read as a text file. Is it Python source?"
                % self.mainpyfile
            )
            self.intf[0].errmsg(sys.exc_info()[1])
            retval = False
            pass
        except IOError:
            print(sys.exc_info()[1])
        except DebuggerQuit:
            retval = False
            pass
        except DebuggerRestart:
            self.core.execution_status = "Restart requested"
            raise DebuggerRestart
        else:
            self.core.start(start_opts)
            exec(compiled, globals_, locals_)
            retval = True
        finally:
            self.core.stop(options={"remove": True})
        return retval

    def restart_argv(self):
        """Return an array that would be execv-ed  to restart the program"""
        return self.orig_sys_argv or self.program_sys_argv

    # The following functions have to be defined before
    # DEFAULT_INIT_OPTS which includes references to these.

    # Note: has to come after functions listed in ignore_filter.
    DEFAULT_INIT_OPTS = {
        # What routines will we not trace into?
        "ignore_filter": tracefilter.TraceFilter(
            [tracer.start, tracer.stop, run_eval, run_call, run_eval, run_script]
        ),
        # sys.argv when not None contains sys.argv *before* debugger
        # command processing. So sys.argv contains debugger options as
        # well as debugged-program options. These options are used to
        # do a "hard" or execv() restart.
        # program_sys_argv is set by option save_sys_argv and contains
        # sys.argv that we see now which may have debugger options
        # stripped, or it may be that we were not called from a
        # debugger front end but from inside the running
        # program. These options are suitable for a "soft" or
        # exception-handling DebuggerRestart kind of restart.
        "orig_sys_argv": None,
        "save_sys_argv": True,
        # How is I/O for this debugger handled?
        "activate": False,
        "interface": None,
        "input": None,
        "output": None,
        "processor": None,
        # Setting contains lots of debugger settings - a whole file
        # full of them!
        "settings": DEBUGGER_SETTINGS,
        "start_opts": START_OPTS,
        "step_ignore": 0,
        "from_ipython": False,
    }

    pass


# Demo it
if __name__ == "__main__":

    def foo():
        y = 2
        for item in range(2):
            print(f"{item} {y}")
            pass
        return 3

    d = Trepan()
    print(d.settings)
    d.settings["trace"] = True
    d.settings["printset"] = tracer.ALL_EVENTS
    d.core.step_ignore = -1
    print('Issuing: run_eval("1+2")')
    print(d.run_eval("1+2"))
    print('Issuing: run_exec("x=1; y=2")')
    d.run_exec("x=1; y=2")

    print('Issuing: run("3*4")')
    print(d.run("3*4"))
    print('Issuing: run("x=3; y=4")')
    d.run("x=3; y=4")

    print("Issuing: run_call(foo)")
    d.run_call(foo)
    if len(sys.argv) > 1:
        while True:
            try:
                print("started")
                d.core.step_ignore = 0
                d.core.start()
                x = foo()
                for i in range(2):
                    print(f"{(i + 1) * 10}")
                    pass
                d.core.stop()

                def square(n):
                    return n * n

                print("calling: run_call(square,2)")
                d.run_call(square, 2)
            except DebuggerQuit:
                print("That's all Folks!...")
                break
            except DebuggerRestart:
                print("Restarting...")
                pass
            pass
        pass
    pass



"""Unit test for trepan.lib.file"""
import os
import stat
import sys
import tempfile

import pytest

from trepan.lib.file import executable, lookupmodule, readable


def test_lookupmodule():
    m, f = lookupmodule("os.path")
    assert f
    assert m
    m, f = lookupmodule(__file__)
    assert f
    assert m is None
    assert (None, None) == lookupmodule("fafdsafdsa")
    return


@pytest.mark.skipif(sys.platform == "win32", reason="Does not work on MS Windows")
def test_readable():
    assert not readable("fdafdsa")
    i = 1
    for mode, can_read in [
        (stat.S_IRUSR, True),  # 1
        (stat.S_IWUSR, False),  # 2
        (stat.S_IRGRP, True),  # 3
        (stat.S_IWGRP, False),  # 4
        (stat.S_IROTH, True),  # 5
        (stat.S_IWOTH, False),  # 6
    ]:
        f = tempfile.NamedTemporaryFile()
        os.chmod(f.name, mode)
        assert can_read == readable(f.name), f"Test {i} file mode {oct(mode)}"
        f.close()
        i += 1
        pass
    return


@pytest.mark.skipif(sys.platform == "win32", reason="Does not work on MS Windows")
def test_executable():
    i = 1
    for mode, can_read in [
        (stat.S_IRUSR | stat.S_IXUSR, True),  # 1
        (stat.S_IRUSR, False),  # 2
        (stat.S_IWGRP | stat.S_IXGRP, False),  # 3
        (stat.S_IWGRP, False),  # 4
        (stat.S_IRGRP | stat.S_IXGRP, True),  # 5
        (stat.S_IRGRP, False),  # 6
        (stat.S_IWGRP, False),  # 7
        (stat.S_IROTH | stat.S_IXOTH, True),  # 8
    ]:
        f = tempfile.NamedTemporaryFile()
        os.chmod(f.name, mode)

        assert can_read == executable(f.name), f"Test {i} file mode {oct(mode)}"
        f.close()
        i += 1
        pass
    return



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Post-Mortem interface

import inspect
import os
import re
import sys
import traceback

# Our local modules
from trepan import debugger as Mdebugger
from trepan.exception import DebuggerQuit, DebuggerRestart


def get_last_or_frame_exception():
    """Intended to be used going into post mortem routines.  If
    sys.last_traceback is set, we will return that and assume that
    this is what post-mortem will want. If sys.last_traceback has not
    been set, then perhaps we *about* to raise an error and are
    fielding an exception. So assume that sys.exc_info()[2]
    is where we want to look."""

    try:
        if inspect.istraceback(sys.last_traceback):
            # We do have a traceback so prefer that.
            return sys.last_type, sys.last_value, sys.last_traceback
    except AttributeError:
        pass
    return sys.exc_info()


def pm(frameno=1, dbg=None):
    """Set up post-mortem debugging using the last traceback.  But if
    there is no traceback, we'll assume that sys.exc_info() contains
    what we want and frameno is the index location of where we want
    to start.

    'dbg', is an optional trepan.Trepan object.
    """
    post_mortem(get_last_or_frame_exception(), frameno, dbg=dbg)
    return


def post_mortem_excepthook(exc_type, exc_value, exc_tb, tb_fn=None):
    if str(exc_type) == str(DebuggerQuit):
        return
    try:
        if str(exc_type) == str(DebuggerRestart):
            if exc_value and exc_value.sys_argv and len(exc_value.sys_argv) > 0:
                print(
                    "No restart handler - trying restart via execv(%s)"
                    % repr(exc_value.sys_argv)
                )
                os.execvp(exc_value.sys_argv[0], exc_value.sys_argv)
            else:
                print("No restart handler, no params registered")
                print("Entering post-mortem debugger...")
        else:
            if tb_fn:
                tb_fn(exc_type, exc_value, exc_tb)
            else:
                traceback.print_exception(exc_type, exc_value, exc_tb)
            print("Uncaught exception. Entering post-mortem debugger...")
            pass
        post_mortem((exc_type, exc_value, exc_tb))
        print("Post-mortem debugger finished.")
    except Exception:
        pass
    return


def post_mortem(exc=None, frameno=1, dbg=None):
    """Enter debugger read loop after your program has crashed.

    exc is a triple like you get back from sys.exc_info.  If no exc
    parameter, is supplied, the values from sys.last_type,
    sys.last_value, sys.last_traceback are used. And if these don't
    exist either we'll assume that sys.exc_info() contains what we
    want and frameno is the index location of where we want to start.

    'frameno' specifies how many frames to ignore in the traceback.
    The default is 1, that is, we don't need to show the immediate
    call into post_mortem. If you have wrapper functions that call
    this one, you may want to increase frameno.
    """

    if dbg is None:
        # Check for a global debugger object
        if Mdebugger.debugger_obj is None:
            Mdebugger.debugger_obj = Mdebugger.Trepan()
            pass
        dbg = Mdebugger.debugger_obj
        pass
    re_bogus_file = re.compile("^<.+>$")

    if exc[0] is None:
        # frameno+1 because we are about to add one more level of call
        # in get_last_or_frame_exception
        exc = get_last_or_frame_exception()
        if exc[0] is None:
            print(
                "Can't find traceback for post_mortem "
                "in sys.last_traceback or sys.exec_info()"
            )
            return
        pass
    exc_type, exc_value, exc_tb = exc
    dbg.core.execution_status = "Terminated with unhandled exception %s" % exc_type

    # tb has least-recent traceback entry first. We want the most-recent
    # entry. Also we'll pick out a mainpyfile name if it hasn't previously
    # been set.
    if exc_tb is not None:
        while exc_tb.tb_next is not None:
            filename = exc_tb.tb_frame.f_code.co_filename
            if (
                dbg.mainpyfile
                and 0 == len(dbg.mainpyfile)
                and not re_bogus_file.match(filename)
            ):
                dbg.mainpyfile = filename
                pass
            exc_tb = exc_tb.tb_next
            pass
        dbg.core.processor.curframe = exc_tb.tb_frame
        pass

    if 0 == len(dbg.program_sys_argv):
        # Fake program (run command) args since we weren't called with any
        dbg.program_sys_argv = list(sys.argv[1:])
        dbg.program_sys_argv[:0] = [dbg.mainpyfile]

    # if 0 == len(dbg._sys_argv):
    #     # Fake script invocation (restart) args since we don't have any
    #     dbg._sys_argv = list(dbg.program_sys_argv)
    #     dbg._sys_argv[:0] = [__title__]

    try:
        # # FIXME: This can be called from except hook in which case we
        # # need this. Dunno why though.
        # try:
        #     _pydb_trace.set_trace(t.tb_frame)
        # except Exception:
        #     pass

        # Possibly a bug in Python 2.5. Why f.f_lineno is
        # not always equal to t.tb_lineno, I don't know.
        f = exc_tb.tb_frame
        if f and f.f_lineno != exc_tb.tb_lineno:
            f = f.f_back
        dbg.core.processor.event_processor(f, "exception", exc, "Trepan3k:pm")
    except DebuggerRestart:
        while True:
            sys.argv = list(dbg._program_sys_argv)
            dbg.msg(
                "Restarting %s with arguments:\n\t%s"
                % (dbg.filename(dbg.mainpyfile), " ".join(dbg._program_sys_argv[1:]))
            )
            try:
                dbg.run_script(dbg.mainpyfile)
            except DebuggerRestart:
                pass
            pass
    except DebuggerQuit:
        pass
    return


def uncaught_exception(dbg, tb_fn=None):
    exc = sys.exc_info()
    exc_type, exc_value, exc_tb = exc
    if exc_type == DebuggerQuit:
        return
    if exc_type == DebuggerRestart:
        print("restart not done yet - entering post mortem debugging")
    elif exc_tb is None:
        print("You don't seem to have an exception traceback, yet.")
        return
    else:
        if tb_fn:
            tb_fn(exc_type, exc_value, exc_tb)
        else:
            traceback.print_exception(exc_type, exc_value, exc_tb)
        print("uncaught exception. entering post mortem debugging")
        pass
    dbg.core.execution_status = "Terminated with unhandled exception %s" % exc_type
    dbg.core.processor.event_processor(exc_tb.tb_frame, "exception", exc, "Trepan3k:pm")
    print("Post mortem debugger finished.")
    return


if __name__ == "__main__":
    if len(sys.argv) > 1:
        pm()
        pass
    pass



"""Unit test for trepan.lib.display"""

import inspect

from trepan.lib.display import Display, DisplayMgr


def test_display():
    mgr = DisplayMgr()
    assert mgr.list == [], "Initial list empty"
    x = 1  # NOQA
    frame = inspect.currentframe()
    disp = mgr.add(frame, "x > 1")
    assert isinstance(disp, Display), "mgr.add should return display"
    assert len(mgr.list) == 1, "display list with one item"
    disp = mgr.add(frame, "x")
    assert isinstance(disp, Display), "mgr.add should return another display"
    assert len(mgr.list) == 2, "display list with two items"
    assert mgr.delete_index(1) is True, "return True on ok delete"
    assert mgr.delete_index(1) is False, "return False on no delete"
    assert len(mgr.list) == 1, "display list again with one item"
    return



#!/usr/bin/env python
"""Unit test for debugger command completion"""

from test.unit.cmdhelper import setup_unit_test_debugger

import pytest
from xdis import IS_PYPY

from trepan.processor import complete as module_complete
from trepan.processor.command import base_cmd as module_base_command

line_buffer = ""


def get_line_buffer():
    return line_buffer


def run_complete(dbgr, line_str: str):
    global line_buffer
    line_buffer = line_str
    results = []
    i = 0
    # from trepan.api import debug; debug()
    got = dbgr.complete(line_str, i)
    while got:
        results.append(got)
        i += 1
        got = dbgr.complete(line_str, i)
        pass
    return results


def test_complete_identifier():
    _, cmdproc = setup_unit_test_debugger()
    cmd = module_base_command.DebuggerCommand(cmdproc)

    assert module_complete.complete_id_and_builtins(cmd, "ma") == ["map", "max"]
    assert module_complete.complete_identifier(cmd, "m") == [
        "module_base_command",
        "module_complete",
    ]


@pytest.mark.skipif(IS_PYPY, reason="Does not work with PyPy")
def test_completion():
    dbgr, _ = setup_unit_test_debugger()

    for line, expect_completion in [
        ["set basename ", ["off", "on"]],
        ["where", ["where "]],  # Single alias completion
        ["sho", ["show"]],  # Simple single completion
        ["un", ["unalias", "undisplay"]],  # Simple multiple completion
        ["python ", []],  # Don't add anything - no more
        ["set basename o", ["off", "on"]],
        ["set basename of", ["off"]],
        # Multiple completion on two words
        ["set auto", ["autoeval", "autolist", "autopc", "autopython"]],
        # Completion when word is complete, without space.
        ["show", ["show "]],
        # Completion when word is complete with space.
        [
            "info ",
            [
                "args",
                "break",
                "builtins",
                "code",
                "display",
                "files",
                "frame",
                "globals",
                "line",
                "lines",
                "locals",
                "macro",
                "offsets",
                "pc",
                "program",
                "return",
                "signals",
                "source",
                "threads",
            ],
        ],
        ["help sta", ["stack", "status"]],
        [" unalias c", ["c", "chdir", "cond"]],
        # Any set style completion
        ["set style def", ["default"]],
        # ['set auto eval ', '', ['off', 'on']],
        # Many 3-word completions
        # ['set auto ', ['eval', 'irb', 'list']],
        # Many two-word completions
        # ['set auto e', ['eval']],
        # ['disas', ['disassemble']], # Another single completion
        # ['help syn', ['syntax']],
        # ## FIXME:
        # ## ['help syntax co', ['command']],
        # ['help br', ['break', 'breakpoints']],
    ]:
        got = run_complete(dbgr, line)
        assert expect_completion == got, "Completion of '%s', expecting %s, got %s" % (
            line,
            expect_completion,
            got,
        )
        pass

    got = run_complete(dbgr, "")
    assert len(got) > 30, "Initial completion should return more than 30 commands"
    got = run_complete(dbgr, "info files ")
    assert len(got) > 0, "info files completion should return a file"
    got = run_complete(dbgr, "unalias ")
    assert len(got) > 0, "unalias should return lots of aliases"
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2010, 2013-2021, 2023-2024 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import importlib
import inspect
import linecache
import os.path as osp
import re
import shlex
import sys
import tempfile
import traceback

# Note: the module name pre 3.2 is repr
from reprlib import Repr

import pyficache
from pygments.console import colorize
from tracer import EVENT2SHORT

import trepan.exception as Mexcept
import trepan.lib.display as Mdisplay
import trepan.lib.file as Mfile
import trepan.lib.stack as Mstack
import trepan.lib.thred as Mthread
import trepan.misc as Mmisc
import trepan.processor.complete as Mcomplete
from trepan.lib.bytecode import is_class_def, is_def_stmt
from trepan.processor import cmdfns
from trepan.processor.cmdfns import deparse_fn
from trepan.vprocessor import Processor

try:
    from trepan.lib.deparse import deparse_and_cache

    have_deparse_and_cache = True
except ImportError:
    have_deparse_and_cache = False
    pass

warned_file_mismatches = set()


def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)


# arg_split culled from ipython's routine
def arg_split(s, posix=False):
    """Split a command line's arguments in a shell-like manner returned
    as a list of lists. Use ';;' with white space to indicate separate
    commands.

    This is a modified version of the standard library's shlex.split()
    function, but with a default of posix=False for splitting, so that quotes
    in inputs are respected.
    """

    args_list = [[]]
    if isinstance(s, bytes):
        s = s.decode("utf-8")
    lex = shlex.shlex(s, posix=posix)

    lex.whitespace_split = True
    args = list(lex)
    for arg in args:
        if ";;" == arg:
            args_list.append([])
        else:
            args_list[-1].append(arg)
            pass
        pass
    return args_list


def get_stack(f, t, botframe, proc_obj=None):
    """Return a stack of frames which the debugger will use for in
    showing backtraces and in frame switching. As such various frame
    that are really around may be excluded unless we are debugging the
    sebugger. Also we will add traceback frame on top if that
    exists."""

    def false_fn(f):
        return false_fn

    def fn_is_ignored(f):
        return proc_obj.core.ignore_filter.is_included(f)

    exclude_frame = false_fn
    if proc_obj:
        settings = proc_obj.debugger.settings
        if not settings["dbg_trepan"]:
            exclude_frame = fn_is_ignored
            pass
        pass
    stack = []
    if t and t.tb_frame is f:
        t = t.tb_next
    while f is not None:
        if exclude_frame(f):
            break  # See commented alternative below
        stack.append((f, f.f_lineno))
        # bdb has:
        # if f is botframe: break
        f = f.f_back
        pass
    stack.reverse()
    i = max(0, len(stack) - 1)
    while t is not None:
        stack.append((t.tb_frame, t.tb_lineno))
        t = t.tb_next
        pass
    return stack, i


def run_hooks(obj, hooks, *args):
    """Run each function in `hooks' with args"""
    for hook in hooks:
        if hook(obj, *args):
            return True
        pass
    return False


def resolve_name(obj, command_name):
    if command_name.lower() not in obj.commands:
        if command_name in obj.aliases:
            command_name = obj.aliases[command_name]
            pass
        else:
            return None
        pass
    try:
        return command_name.lower()
    except Exception:
        return None
    return


def print_source_line(msg, lineno, line, event_str=None):
    """Print out a source line of text , e.g. the second
    line in:
        (/tmp.py:2):  <module>
        L -- 2 import sys,os
        (trepan3k)

    We define this method
    specifically so it can be customized for such applications
    like ipython."""

    # We don't use the filename normally. ipython and other applications
    # however might.
    return msg("%s %d %s" % (event_str, lineno, line))


def print_source_location_info(
    print_fn, filename, lineno, fn_name=None, f_lasti=None, remapped_file=None
):
    """Print out a source location , e.g. the first line in
    line in:
        (/tmp.py:2 @21):  <module>
        L -- 2 import sys,os
        (trepan3k)
    """
    if remapped_file:
        mess = f"({remapped_file}:{lineno} remapped {filename}"
    else:
        mess = f"({filename}:{lineno}"
    if f_lasti and f_lasti != -1:
        mess += " @%d" % f_lasti
        pass
    mess += "):"
    if fn_name and fn_name != "?":
        mess += f" {fn_name}"
        pass
    print_fn(mess)
    return


def print_location(proc_obj):
    """Show where we are. GUI's and front-end interfaces often
    use this to update displays. So it is helpful to make sure
    we give at least some place that's located in a file.
    """
    i_stack = proc_obj.curindex
    if i_stack is None or proc_obj.stack is None:
        return False
    core_obj = proc_obj.core
    dbgr_obj = proc_obj.debugger
    intf_obj = dbgr_obj.intf[-1]

    # Evaluation routines like "exec" don't show useful location
    # info. In these cases, we will use the position before that in
    # the stack.  Hence the looping below which in practices loops
    # once and sometimes twice.
    remapped_file = None
    source_text = None
    while i_stack >= 0:
        frame, lineno = proc_obj.stack[i_stack]

        # Before starting a program a location for a module with
        # line number 0 may be reported. Treat that as though
        # we were on the first line.
        if frame.f_code.co_name == "<module>" and lineno == 0:
            lineno = 1

        i_stack -= 1

        #         # Next check to see that local variable breadcrumb exists and
        #         # has the magic dynamic value.
        #         # If so, it's us and we don't normally show this.a
        #         if 'breadcrumb' in frame.f_locals:
        #             if self.run == frame.f_locals['breadcrumb']:
        #                 break

        filename = Mstack.frame2file(core_obj, frame, canonic=False)
        if "<string>" == filename and dbgr_obj.eval_string:
            remapped_file = filename
            filename = pyficache.unmap_file(filename)
            if "<string>" == filename:
                remapped = cmdfns.source_tempfile_remap(
                    "eval_string",
                    dbgr_obj.eval_string,
                    tempdir=proc_obj.settings("tempdir"),
                )
                pyficache.remap_file(filename, remapped)
                filename, lineno = pyficache.unmap_file_line(filename, lineno)
                pass
            pass
        elif "<string>" == filename:
            source_text = deparse_fn(frame.f_code)
            filename = f"<string: '{source_text}'>"
            pass
        else:
            m = re.search("^<frozen (.*)>", filename)
            if m and m.group(1) in pyficache.file2file_remap:
                remapped_file = pyficache.file2file_remap[m.group(1)]
                pass
            elif filename in pyficache.file2file_remap:
                remapped_file = pyficache.unmap_file(filename)
                # FIXME: a remapped_file shouldn't be the same as its unmapped version
                if remapped_file == filename:
                    remapped_file = None
                    pass
                pass
            elif pyficache.main.remap_re_hash:
                remapped_file = pyficache.remap_file_pat(
                    filename, pyficache.main.remap_re_hash
                )
            elif m and m.group(1) in sys.modules:
                remapped_file = m.group(1)
                pyficache.remap_file(filename, remapped_file)
            pass

        opts = {
            "reload_on_change": proc_obj.settings("reload"),
            "output": proc_obj.settings("highlight"),
        }

        if "style" in proc_obj.debugger.settings:
            opts["style"] = proc_obj.settings("style")

        pyficache.update_cache(filename)
        line = pyficache.getline(filename, lineno, opts)
        if not line:
            if (
                not source_text
                and filename.startswith("<string: ")
                and proc_obj.curframe.f_code
                and have_deparse_and_cache
            ):
                # Deparse the code object into a temp file and remap the line from code
                # into the corresponding line of the tempfile
                co = proc_obj.curframe.f_code
                tempdir = proc_obj.settings("tempdir")
                temp_filename, name_for_code = deparse_and_cache(
                    co, proc_obj.errmsg, tempdir=tempdir
                )
                lineno = 1
                # _, lineno = pyficache.unmap_file_line(temp_filename, lineno, True)
                if temp_filename:
                    filename = temp_filename
                pass

            else:
                # FIXME:
                if source_text:
                    lines = source_text.split("\n")
                    temp_name = "string-"
                else:
                    # try with good ol linecache and consider fixing pyficache
                    lines = linecache.getlines(filename)
                    temp_name = filename
                if lines:
                    # FIXME: DRY code with version in cmdproc.py print_location
                    prefix = osp.basename(temp_name).split(".")[0]
                    fd = tempfile.NamedTemporaryFile(
                        suffix=".py",
                        prefix=prefix,
                        delete=False,
                        dir=proc_obj.settings("tempdir"),
                    )
                    with fd:
                        fd.write("".join(lines).encode("utf-8"))
                        remapped_file = fd.name
                        pyficache.remap_file(remapped_file, filename)
                    fd.close()
                    intf_obj.msg(f"remapped file {filename} to {remapped_file}")

                    pass
            line = linecache.getline(filename, lineno, proc_obj.curframe.f_globals)
            if not line:
                m = re.search("^<frozen (.*)>", filename)
                if m and m.group(1):
                    remapped_file = m.group(1)
                    try_module = sys.modules.get(remapped_file)
                    if (
                        try_module
                        and inspect.ismodule(try_module)
                        and hasattr(try_module, "__file__")
                    ):
                        remapped_file = sys.modules[remapped_file].__file__
                        pyficache.remap_file(filename, remapped_file)
                        line = linecache.getline(
                            remapped_file, lineno, proc_obj.curframe.f_globals
                        )
                    else:
                        remapped_file = m.group(1)
                        code = proc_obj.curframe.f_code
                        filename, line = cmdfns.deparse_getline(
                            code, remapped_file, lineno, opts
                        )
                    pass
            pass

        try:
            match, reason = Mstack.check_path_with_frame(frame, filename)
            if not match:
                if filename not in warned_file_mismatches:
                    proc_obj.errmsg(reason)
                    warned_file_mismatches.add(filename)
        except Exception:
            pass

        fn_name = frame.f_code.co_name
        last_i = frame.f_lasti
        print_source_location_info(
            intf_obj.msg,
            filename,
            lineno,
            fn_name,
            remapped_file=remapped_file,
            f_lasti=last_i,
        )
        if line and len(line.strip()) != 0:
            if proc_obj.event:
                print_source_line(
                    intf_obj.msg, lineno, line, proc_obj.event2short[proc_obj.event]
                )
            pass
        if "<string>" != filename:
            break
        pass

    if proc_obj.event in ["return", "exception"]:
        val = proc_obj.event_arg
        intf_obj.msg(f"R=> {proc_obj._saferepr(val)}")
        pass
    elif (
        proc_obj.event == "call"
        and proc_obj.curframe.f_locals.get("__name__", "") != "__main__"
    ):
        try:
            proc_obj.commands["info"].run(["info", "locals"])
        except Exception:
            pass
    return True


# Default settings for command processor method call
DEFAULT_PROC_OPTS = {
    # A list of debugger initialization files to read on first command
    # loop entry.  Often this something like [~/.config/trepanpy/profile] which the
    # front-end sets.
    "initfile_list": []
}


class CommandProcessor(Processor):
    def __init__(self, core_obj, opts=None):
        def get_option_fn(key):
            return Mmisc.option_set(opts, key, DEFAULT_PROC_OPTS)

        get_option = get_option_fn
        super().__init__(core_obj)

        self.continue_running = False  # True if we should leave command loop
        self.event2short = dict(EVENT2SHORT)
        self.event2short["signal"] = "?!"
        self.event2short["brkpt"] = "xx"

        self.optional_modules = ("ipython", "bpy")
        self.cmd_instances = self._populate_commands()

        # command argument string. Is like current_command, but the part
        # after cmd_name has been removed.
        self.cmd_argstr = ""

        # command name before alias or macro resolution
        self.cmd_name = ""
        self.cmd_queue = []  # Queued debugger commands
        self.completer = lambda text, state: Mcomplete.completer(self, text, state)
        self.current_command = ""  # Current command getting run
        self.debug_nest = 1
        self.display_mgr = Mdisplay.DisplayMgr()
        self.intf = core_obj.debugger.intf
        self.last_command = None  # Initially a no-op
        self.precmd_hooks = []

        self.location = lambda: print_location(self)

        self.preloop_hooks = []
        self.postcmd_hooks = []
        self.remap_file_re = None

        self._populate_cmd_lists()

        # Note: prompt_str's value set below isn't used. It is
        # computed dynamically. The value is suggestive of what it
        # looks like.
        self.prompt_str = "(trepan3k) "

        # Stop only if line/file is different from last time
        self.different_line = None

        # These values updated on entry. Set initial values.
        self.curframe = None
        self.event = None
        self.event_arg = None
        self.frame = None
        self.list_lineno = 0  # last list number used in "list"
        self.list_offset = -1  # last list number used in "disassemble"
        self.list_obj = None
        self.list_filename = None  # last filename used in list
        self.list_orig_lineno = 0  # line number of frame or exception on setup
        self.list_filename = None  # filename of frame or exception on setup

        self.macros = {}  # Debugger Macros

        # Create a custom safe Repr instance and increase its maxstring.
        # The default of 30 truncates error messages too easily.
        self._repr = Repr()
        self._repr.maxstring = 100
        self._repr.maxother = 60
        self._repr.maxset = 10
        self._repr.maxfrozen = 10
        self._repr.array = 10
        self.stack = []
        self.thread_name = None
        self.frame_thread_name = None

        initfile_list = get_option("initfile_list")
        for init_cmdfile in initfile_list:
            self.queue_startfile(init_cmdfile)
        return

    def _saferepr(self, str, maxwidth=None):
        if maxwidth is None:
            maxwidth = self.debugger.settings["width"]
        return self._repr.repr(str)[:maxwidth]

    def add_preloop_hook(self, hook, position=-1, nodups=True):
        if hook in self.preloop_hooks:
            return False
        self.preloop_hooks.insert(position, hook)
        return True

    def add_remap_pat(self, pat, replace, clear_remap=True):
        pyficache.main.add_remap_pat(pat, replace, clear_remap)
        if clear_remap:
            self.file2file_remap = {}
            pyficache.file2file_remap = {}

    # To be overridden in derived debuggers
    def defaultFile(self):
        """Produce a reasonable default."""
        filename = self.curframe.f_code.co_filename
        # Consider using is_exec_stmt(). I just don't understand
        # the conditions under which the below test is true.
        if filename == "<string>" and self.debugger.mainpyfile:
            filename = self.debugger.mainpyfile
            pass
        return filename

    def set_prompt(self, prompt="trepan3k"):
        if self.thread_name and self.thread_name != "MainThread":
            prompt += ":" + self.thread_name
            pass
        self.prompt_str = f"{'(' * self.debug_nest}{prompt}{')' * self.debug_nest}"
        highlight = self.debugger.settings["highlight"]
        if highlight and highlight in ("light", "dark"):
            self.prompt_str = colorize("underline", self.prompt_str)
        self.prompt_str += " "

    def event_processor(self, frame, event, event_arg, prompt="trepan3k"):
        """
        command event processor: reading a commands do something with them.

        See https://docs.python.org/3/library/sys.html#sys.settrace
        for how this protocol works and what the events means.

        Of particular note those is what we return:

            The local trace function should return a reference to
            itself (or to another function for further tracing in that
            scope), or None to turn off tracing in that scope.

            If there is any error occurred in the trace function, it
            will be unset, just like settrace(None) is called.
        """

        self.frame = frame
        self.event = event
        self.event_arg = event_arg

        filename = frame.f_code.co_filename
        lineno = frame.f_lineno
        line = linecache.getline(filename, lineno, frame.f_globals)
        if not line:
            opts = {
                "output": "plain",
                "reload_on_change": self.settings("reload"),
                "strip_nl": False,
            }
            m = re.search("^<frozen (.*)>", filename)
            if m and m.group(1):
                filename = pyficache.unmap_file(m.group(1))
            line = pyficache.getline(filename, lineno, opts)
        self.current_source_text = line
        if self.settings("skip") is not None:
            if is_def_stmt(line, frame):
                return self.event_processor
            if is_class_def(line, frame):
                return self.event_processor
            pass
        self.thread_name = Mthread.current_thread_name()
        self.frame_thread_name = self.thread_name
        self.set_prompt(prompt)
        self.process_commands()
        if filename == "<string>":
            pyficache.remove_remap_file("<string>")
        return self.event_processor

    def forget(self):
        """Remove memory of state variables set in the command processor"""
        self.stack = []
        self.curindex = 0
        self.curframe = None
        self.thread_name = None
        self.frame_thread_name = None
        return

    def eval(self, arg, show_error=True):
        """Eval string arg in the current frame context."""
        try:
            return eval(arg, self.curframe.f_globals, self.curframe.f_locals)
        except Exception:
            t, v = sys.exc_info()[:2]
            if isinstance(t, str):
                exc_type_name = t
                pass
            else:
                exc_type_name = t.__name__
            if show_error:
                self.errmsg(str(f"{exc_type_name}: {arg}"))
            raise
        return None  # Not reached

    def exec_line(self, line):
        if self.curframe:
            local_vars = self.curframe.f_locals
            global_vars = self.curframe.f_globals
        else:
            local_vars = None
            # FIXME: should probably have place where the
            # user can store variables inside the debug session.
            # The setup for this should be elsewhere. Possibly
            # in interaction.
            global_vars = None
        try:
            code = compile(line + "\n", f'"{line}"', "single")
            exec(code, global_vars, local_vars)
        except Exception:
            t, v = sys.exc_info()[:2]
            if type(t) == bytes:
                exc_type_name = t
            else:
                exc_type_name = t.__name__
            self.errmsg(f"{str(exc_type_name)}: {str(v)}")
            pass
        return

    def get_an_int(self, arg, msg_on_error, min_value=None, max_value=None):
        """Like cmdfns.get_an_int(), but if there's a stack frame use that
        in evaluation."""
        ret_value = self.get_int_noerr(arg)
        if ret_value is None:
            if msg_on_error:
                self.errmsg(msg_on_error)
            else:
                self.errmsg(f"Expecting an integer, got: {str(arg)}.")
                pass
            return None
        if min_value and ret_value < min_value:
            self.errmsg(
                "Expecting integer value to be at least %d, got: %d."
                % (min_value, ret_value)
            )
            return None
        elif max_value and ret_value > max_value:
            self.errmsg(
                "Expecting integer value to be at most %d, got: %d."
                % (max_value, ret_value)
            )
            return None
        return ret_value

    def get_int_noerr(self, arg):
        """Eval arg and it is an integer return the value. Otherwise
        return None"""
        if self.curframe:
            g = self.curframe.f_globals
            locals_dict = self.curframe.f_locals
        else:
            g = globals()
            locals_dict = locals()
            pass
        try:
            val = int(eval(arg, g, locals_dict))
        except (SyntaxError, NameError, ValueError, TypeError):
            return None
        return val

    def get_int(self, arg, min_value=0, default=1, cmdname=None, at_most=None):
        """If no argument use the default. If arg is a an integer between
        least min_value and at_most, use that. Otherwise report an error.
        If there's a stack frame use that in evaluation."""

        if arg is None:
            return default
        default = self.get_int_noerr(arg)
        if default is None:
            if cmdname:
                self.errmsg(
                    ("Command '%s' expects an integer; " + "got: %s.")
                    % (cmdname, str(arg))
                )
            else:
                self.errmsg(f"Expecting a positive integer, got: {str(arg)}")
                pass
            return None
            pass
        if default < min_value:
            if cmdname:
                self.errmsg(
                    ("Command '%s' expects an integer at least" + " %d; got: %d.")
                    % (cmdname, min_value, default)
                )
            else:
                self.errmsg(
                    ("Expecting a positive integer at least" + " %d; got: %d")
                    % (min_value, default)
                )
                pass
            return None
        elif at_most and default > at_most:
            if cmdname:
                self.errmsg(
                    ("Command '%s' expects an integer at most" + " %d; got: %d.")
                    % (cmdname, at_most, default)
                )
            else:
                self.errmsg(
                    ("Expecting an integer at most %d; got: %d") % (at_most, default)
                )
                pass
            pass
        return default

    def getval(self, arg, locals=None):
        if not locals:
            locals = self.curframe.f_locals
        try:
            return eval(arg, self.curframe.f_globals, locals)
        except Exception:
            t, v = sys.exc_info()[:2]
            if isinstance(t, str):
                exc_type_name = t
            else:
                exc_type_name = t.__name__
            self.errmsg(str(f"{exc_type_name}: {arg}"))
            raise
        return

    def ok_for_running(self, cmd_obj, name, nargs):
        """We separate some of the common debugger command checks here:
        whether it makes sense to run the command in this execution state,
        if the command has the right number of arguments and so on.
        """
        if hasattr(cmd_obj, "execution_set"):
            if not (self.core.execution_status in cmd_obj.execution_set):
                part1 = f"Command '{name}' is not available for execution status:"
                mess = Mmisc.wrapped_lines(
                    part1, self.core.execution_status, self.debugger.settings["width"]
                )
                self.errmsg(mess)
                return False
            pass
        if self.frame is None and cmd_obj.need_stack:
            self.intf[-1].errmsg(f"Command '{name}' needs an execution stack.")
            return False
        if nargs < cmd_obj.min_args:
            self.errmsg(
                ("Command '%s' needs at least %d argument(s); " + "got %d.")
                % (name, cmd_obj.min_args, nargs)
            )
            return False
        elif cmd_obj.max_args is not None and nargs > cmd_obj.max_args:
            self.errmsg(
                ("Command '%s' can take at most %d argument(s);" + " got %d.")
                % (name, cmd_obj.max_args, nargs)
            )
            return False
        return True

    def process_commands(self):
        """Handle debugger commands."""
        if self.core.execution_status != "No program":
            self.setup()
            self.location()
            pass
        else:
            self.list_object = None

        leave_loop = run_hooks(self, self.preloop_hooks)
        self.continue_running = False

        while not leave_loop:
            try:
                run_hooks(self, self.precmd_hooks)
                # bdb had a True return to leave loop.
                # A more straight-forward way is to set
                # instance variable self.continue_running.
                leave_loop = self.process_command()
                if leave_loop or self.continue_running:
                    break
            except EOFError:
                # If we have stacked interfaces, pop to the next
                # one.  If this is the last one however, we'll
                # just stick with that.  FIXME: Possibly we should
                # check to see if we are interactive.  and not
                # leave if that's the case. Is this the right
                # thing?  investigate and fix.
                if len(self.debugger.intf) > 1:
                    del self.debugger.intf[-1]
                    self.last_command = ""
                else:
                    if self.debugger.intf[-1].output:
                        self.debugger.intf[-1].output.writeline("Leaving")
                        raise SystemExit
                        pass
                    break
                pass
            pass
        return run_hooks(self, self.postcmd_hooks)

    def process_command(self):
        # process command
        if len(self.cmd_queue) > 0:
            current_command = self.cmd_queue[0].strip()
            del self.cmd_queue[0]
        else:
            current_command = self.intf[-1].read_command(self.prompt_str).strip()
            if "" == current_command and self.intf[-1].interactive:
                current_command = self.last_command
                pass
            pass
        # Look for comments
        if "" == current_command:
            if self.intf[-1].interactive:
                self.errmsg("No previous command registered, " + "so this is a no-op.")
                pass
            return False
        if current_command is None or current_command[0] == "#":
            return False
        try:
            args_list = arg_split(current_command)
        except Exception:
            self.errmsg("bad parse %s: %s" % sys.exc_info()[0:2])
            return False

        for args in args_list:
            if len(args):
                while True:
                    if len(args) == 0:
                        return False
                    macro_cmd_name = args[0]
                    if macro_cmd_name not in self.macros:
                        break
                    try:
                        current_command = self.macros[macro_cmd_name][0](*args[1:])
                    except TypeError:
                        t, v = sys.exc_info()[:2]
                        self.errmsg(f"Error expanding macro {macro_cmd_name}")
                        return False
                    if self.settings("debugmacro"):
                        print(current_command)
                        pass
                    if isinstance(current_command, list):
                        for x in current_command:
                            if str != type(x):
                                self.errmsg(
                                    (
                                        "macro %s should return a List "
                                        + "of Strings. Has %s of type %s"
                                    )
                                    % (
                                        macro_cmd_name,
                                        x,
                                        repr(current_command),
                                        type(x),
                                    )
                                )
                                return False
                            pass

                        first = current_command[0]
                        args = first.split()
                        self.cmd_queue + [current_command[1:]]
                        current_command = first
                    elif type(current_command) == str:
                        args = current_command.split()
                    else:
                        self.errmsg(
                            (
                                "macro %s should return a List "
                                + "of Strings or a String. Got %s"
                            )
                            % (macro_cmd_name, repr(current_command))
                        )
                        return False
                    pass

                self.cmd_name = args[0]
                cmd_name = resolve_name(self, self.cmd_name)
                self.cmd_argstr = current_command[len(self.cmd_name) :].lstrip()
                if cmd_name:
                    self.last_command = current_command
                    cmd_obj = self.commands[cmd_name]
                    if self.ok_for_running(cmd_obj, cmd_name, len(args) - 1):
                        try:
                            self.current_command = current_command
                            result = cmd_obj.run(args)
                            if result:
                                return result
                        except (
                            Mexcept.DebuggerQuit,
                            Mexcept.DebuggerRestart,
                            SystemExit,
                        ):
                            # Let these exceptions propagate through
                            raise
                        except Exception:
                            self.errmsg("INTERNAL ERROR: " + traceback.format_exc())
                            pass
                        pass
                    pass
                elif not self.settings("autoeval"):
                    self.undefined_cmd(current_command)
                else:
                    # Autoeval
                    self._saferepr(self.exec_line(current_command))
                    pass
                pass
            pass
        return False

    def remove_preloop_hook(self, hook):
        try:
            position = self.preloop_hooks.index(hook)
        except ValueError:
            return False
        del self.preloop_hooks[position]
        return True

    def setup(self):
        """Initialization done before entering the debugger-command
        loop. In particular we set up the call stack used for local
        variable lookup and frame/up/down commands.

        We return True if we should NOT enter the debugger-command
        loop."""
        self.forget()
        if self.settings("dbg_trepan"):
            self.frame = inspect.currentframe()
            pass
        if self.event in ["exception", "c_exception"]:
            exc_type, exc_value, exc_traceback = self.event_arg
        else:
            _, _, exc_traceback = (
                None,
                None,
                None,
            )  # NOQA
            pass
        if self.frame or exc_traceback:
            self.stack, self.curindex = get_stack(self.frame, exc_traceback, None, self)
            self.curframe = self.stack[self.curindex][0]
            self.thread_name = Mthread.current_thread_name()
            if exc_traceback:
                self.list_lineno = traceback.extract_tb(exc_traceback, 1)[0][1]
                self.list_offset = self.curframe.f_lasti
                self.list_object = self.curframe
        else:
            self.stack = self.curframe = self.botframe = None
            pass
        if self.curframe:
            self.list_lineno = (
                max(
                    1,
                    inspect.getlineno(self.curframe)
                    - int(self.settings("listsize") / 2),
                )
                - 1
            )
            self.list_offset = self.curframe.f_lasti
            self.list_filename = self.curframe.f_code.co_filename
            self.list_object = self.curframe
        else:
            self.list_object = None
            if not exc_traceback:
                self.list_lineno = None
            pass
        # if self.execRcLines()==1: return True

        # FIXME:  do we want to save self.list_lineno a second place
        # so that we can do 'list .' and go back to the first place we listed?
        return False

    def queue_startfile(self, cmdfile):
        """Arrange for file of debugger commands to get read in the
        process-command loop."""
        expanded_cmdfile = osp.expanduser(cmdfile)
        is_readable = Mfile.readable(expanded_cmdfile)
        if is_readable:
            self.cmd_queue.append("source " + expanded_cmdfile)
        elif is_readable is None:
            self.errmsg(f"source file '{expanded_cmdfile}' doesn't exist")
        else:
            self.errmsg(f"source file '{expanded_cmdfile}' is not readable")
            pass
        return

    def undefined_cmd(self, cmd):
        """Error message when a command doesn't exist"""
        self.errmsg(f'Undefined command: "{cmd}". Try "help".')
        return

    def read_history_file(self):
        """Read the command history file -- possibly."""
        histfile = self.debugger.intf[-1].histfile
        try:
            import readline

            readline.read_history_file(histfile)
        except IOError:
            pass
        except ImportError:
            pass
        return

    def write_history_file(self):
        """Write the command history file -- possibly."""
        settings = self.debugger.settings
        histfile = self.debugger.intf[-1].histfile
        if settings["hist_save"]:
            try:
                import readline

                try:
                    readline.write_history_file(histfile)
                except IOError:
                    pass
            except ImportError:
                pass
            pass
        return

    def _populate_commands(self):
        """Create an instance of each of the debugger
        commands. Commands are found by importing files in the
        directory 'command'. Some files are excluded via an array set
        in __init__.  For each of the remaining files, we import them
        and scan for class names inside those files and for each class
        name, we will create an instance of that class. The set of
        DebuggerCommand class instances form set of possible debugger
        commands."""
        from trepan.processor import command as Mcommand

        if hasattr(Mcommand, "__modules__"):
            return self.populate_commands_easy_install(Mcommand)
        else:
            return self.populate_commands_pip(Mcommand, "trepan")

    def populate_commands_pip(self, Mcommand):
        cmd_instances = []
        eval_cmd_template = "command_mod.%s(self)"
        for mod_name in Mcommand.__dict__.keys():
            if mod_name.startswith("__"):
                continue
            import_name = "trepan.processor.command." + mod_name
            imp = __import__(import_name)
            if imp.__name__ == mod_name:
                command_mod = imp.processor.command
            else:
                if mod_name in (
                    "info_sub",
                    "set_sub",
                    "show_sub",
                ):
                    pass
                try:
                    command_mod = getattr(__import__(import_name), mod_name)
                except Exception:
                    # Don't need to warn about optional modules
                    if mod_name not in self.optional_modules:
                        print(f"Error importing {mod_name}: {sys.exc_info()[0]}")
                        pass
                    continue
                pass

            classnames = [
                tup[0]
                for tup in inspect.getmembers(command_mod, inspect.isclass)
                if ("DebuggerCommand" != tup[0] and tup[0].endswith("Command"))
            ]
            for classname in classnames:
                eval_cmd = eval_cmd_template % classname
                if False:
                    instance = eval(eval_cmd)
                    cmd_instances.append(instance)
                else:
                    try:
                        instance = eval(eval_cmd)
                        cmd_instances.append(instance)
                    except Exception:
                        print(
                            "Error loading %s from %s: %s"
                            % (classname, mod_name, sys.exc_info()[0])
                        )
                        pass
                    pass
                pass
            pass
        return cmd_instances

    # This is the most-used way of adding commands
    def populate_commands_easy_install(self, Mcommand):
        """
        Add files in filesystem to self.commands.
        If running from source or from an easy_install'd package, this is used.
        """
        cmd_instances = []

        for mod_name in Mcommand.__modules__:
            if mod_name in (
                "info_sub",
                "set_sub",
                "show_sub",
            ):
                pass
            import_name = f"{Mcommand.__name__}.{mod_name}"
            try:
                command_mod = importlib.import_module(import_name)
            except Exception:
                if mod_name not in self.optional_modules:
                    print(f"Error importing {mod_name}: {sys.exc_info()[0]}")
                    pass
                continue

            classnames = [
                tup[0]
                for tup in inspect.getmembers(command_mod, inspect.isclass)
                if ("DebuggerCommand" != tup[0] and tup[0].endswith("Command"))
            ]
            for classname in classnames:
                if False:
                    instance = getattr(command_mod, classname)(self)
                    cmd_instances.append(instance)
                else:
                    try:
                        instance = getattr(command_mod, classname)(self)
                        cmd_instances.append(instance)
                    except Exception:
                        print(
                            "Error loading %s from %s: %s"
                            % (classname, mod_name, sys.exc_info()[0])
                        )
                        pass
                    pass
                pass
            pass
        return cmd_instances

    def _populate_cmd_lists(self):
        """Populate self.lists and hashes:
        self.commands, and self.aliases, self.category"""
        self.commands = {}
        self.aliases = {}
        self.category = {}
        #         self.short_help = {}
        for cmd_instance in self.cmd_instances:
            if not hasattr(cmd_instance, "aliases"):
                continue
            alias_names = cmd_instance.aliases
            cmd_name = cmd_instance.name
            self.commands[cmd_name] = cmd_instance
            for alias_name in alias_names:
                self.aliases[alias_name] = cmd_name
                pass
            cat = getattr(cmd_instance, "category")
            if cat and self.category.get(cat):
                self.category[cat].append(cmd_name)
            else:
                self.category[cat] = [cmd_name]
                pass
            #             sh = getattr(cmd_instance, 'short_help')
            #             if sh:
            #                 self.short_help[cmd_name] = getattr(c, 'short_help')
            #                 pass
            pass
        for k in list(self.category.keys()):
            self.category[k].sort()
            pass

        return

    pass


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    d = Mmock.MockDebugger()
    cmdproc = CommandProcessor(d.core)
    print("commands:")
    commands = list(cmdproc.commands.keys())
    commands.sort()
    print(commands)
    print("aliases:")
    aliases = list(cmdproc.aliases.keys())
    aliases.sort()
    print(aliases)
    print(resolve_name(cmdproc, "quit"))
    print(resolve_name(cmdproc, "q"))
    print(resolve_name(cmdproc, "info"))
    print(resolve_name(cmdproc, "i"))
    # print '-' * 10
    # print_source_line(sys.stdout.write, 100, 'source_line_test.py')
    # print '-' * 10
    cmdproc.frame = sys._getframe()
    cmdproc.setup()
    print()
    print("-" * 10)
    cmdproc.location()
    print("-" * 10)
    print(cmdproc.eval("1+2"))
    print(cmdproc.eval("len(aliases)"))
    import pprint

    print(pprint.pformat(cmdproc.category))
    print(arg_split("Now is the time"))
    print(arg_split("Now is the time ;;"))
    print(arg_split("Now is 'the time'"))
    print(arg_split("Now is the time ;; for all good men"))
    print(arg_split("Now is the time ';;' for all good men"))

    print(cmdproc.commands)
    fn = cmdproc.commands["quit"]

    print(f"Removing non-existing quit hook: {cmdproc.remove_preloop_hook(fn)}")
    cmdproc.add_preloop_hook(fn)
    print(cmdproc.preloop_hooks)
    print(f"Removed existing quit hook: {cmdproc.remove_preloop_hook(fn)}")
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.cmdfns import want_different_line


class NextCommand(DebuggerCommand):
    """**next**[**+**|**-**] [*count*]

    Execute the current simple statement stopping at the next event but
    ignoring steps into function calls at this level,

    With an integer argument, perform `next` that many times. However if
    an exception occurs at this level, or we *return*, *yield* or the
    thread changes, we stop regardless of count.

    A suffix of `+` on the command or an alias to the command forces to
    move to another line, while a suffix of `-` does the opposite and
    disables the requiring a move to a new line. If no suffix is given,
    the debugger setting 'different-line' determines this behavior.

    See also:
    ---------

    `step`, `skip`, `jump` (there's no `hop` yet), `continue`, and
    `finish` for other ways to progress execution."""

    aliases = ("next+", "next-", "n", "n-", "n+")
    execution_set = ["Running"]
    short_help = "Step over"

    DebuggerCommand.setup(locals(), category="running", need_stack=True, max_args=1)

    def run(self, args):
        if len(args) <= 1:
            step_ignore = 0
        else:
            step_ignore = self.proc.get_int(args[1], default=1, cmdname="next")
            if step_ignore is None:
                return False
            # 0 means stop now or step 1, so we subtract 1.
            step_ignore -= 1
            pass
        self.core.different_line = want_different_line(
            args[0], self.debugger.settings["different"]
        )
        self.core.set_next(self.proc.frame, step_ignore)
        self.proc.continue_running = True  # Break out of command read loop
        return True

    pass


if __name__ == "__main__":
    from mock import MockDebugger

    d = MockDebugger()
    cmd = NextCommand(d.core.processor)
    for c in (["n", "5"], ["next", "1+2"], ["n", "foo"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print("Run result: %s" % result)
        print(
            "step_ignore %d, continue_running: %s"
            % (
                d.core.step_ignore,
                cmd.continue_running,
            )
        )
        pass
    for c in (["n"], ["next+"], ["n-"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print(cmd.core.different_line)
        pass
    pass



#  Copyright (c) 2017-2018, 2020 by Rocky Bernstein

from trepan.processor.parse.parser import (
    parse_bp_location,
    parse_range,
    parse_arange,
    parse_location,
)
from trepan.processor.parse.parser import LocationError as PLocationError
from trepan.processor.parse.scanner import ScannerError
from spark_parser import GenericASTTraversal  # , DEFAULT_DEBUG as PARSER_DEFAULT_DEBUG

from collections import namedtuple

Location = namedtuple("Location", "path line_number is_address method offset")
BPLocation = namedtuple("BPLocation", "location condition")
ListRange = namedtuple("ListRange", "first last")


class LocationError(Exception):
    def __init__(self, errmsg):
        self.errmsg = errmsg

    def __str__(self):
        return self.errmsg


class RangeError(Exception):
    def __init__(self, errmsg):
        self.errmsg = errmsg

    def __str__(self):
        return self.errmsg


class LocationGrok(GenericASTTraversal, object):
    def __init__(self, text):
        GenericASTTraversal.__init__(self, None)
        self.text = text
        self.result = None
        return

    def n_addr_location(self, node):
        # addr_location ::= location
        # addr_location ::= ADDRESS
        # addr_location ::= FUNCNAME (via location singleton reduce)
        # addr_location ::= NUMBER (via location singleton reduce)

        path, line_number, method = None, None, None
        n0 = node[0]
        if n0 in ("ADDRESS", "FUNCNAME", "NUMBER"):
            if n0 == "ADDRESS":
                assert node[0].value[0] == "*"
                line_number = int(node[0].value[1:])
                self.result = Location(path, line_number, True, method, offset=None)
            elif n0 == "NUMBER":
                self.result = Location(path, node[0].value, True, method, offset=None)
            else:
                self.result = Location(
                    path, line_number, False, node[0].value[:-2], offset=None
                )
            node.location = self.result
            self.prune()
        elif n0 == "OFFSET":
            print("WOOT")
        else:
            # print(node[0])
            assert node[0] == "location"
            self.preorder(node[0])
            node.location = node[0].location

    def n_location(self, node):
        path, line_number, method, offset = None, None, None, None
        if node[0] == "FILENAME":
            path = node[0].value
            # If there is a line number, it is the last token of a location
            if len(node) > 1 and node[-1] == "NUMBER":
                line_number = node[-1].value
        elif node[0] == "FUNCNAME":
            method = node[0].value[:-2]
            offset = 0
        elif node[0] == "NUMBER":
            line_number = node[0].value
        else:
            assert True, "n_location: Something's is wrong; node[0] is %s" % node[0]
        self.result = Location(path, line_number, False, method, offset=offset)
        node.location = Location(path, line_number, False, method, offset=offset)
        self.prune()

    def n_NUMBER(self, node):
        self.result = Location(None, node.value, False, None, offset=None)

    def n_ADDRESS(self, node):
        self.result = Location(None, None, True, None, offset=int(node.value[1:]))

    def n_FUNCNAME(self, node):
        self.result = Location(None, None, False, node.value[:-2], offset=0)

    def n_location_if(self, node):
        location = None
        if node[0] == "location":
            self.preorder(node[0])
            location = node[0].location

        if len(node) == 1:
            return
        if node[1] == "IF":
            if_node = node[1]
        elif node[2] == "IF":
            if_node = node[2]
        elif node[3] == "IF":
            if_node = node[3]
        else:
            assert False, 'location_if: Something is wrong; cannot find "if"'

        condition = self.text[if_node.offset + len(if_node.value) + 1 :]

        # Pick out condition from string and location inside "IF" token
        self.result = BPLocation(location, condition)
        self.prune()

    # FIXME: DRY with range
    def n_arange(self, arange_node):
        if arange_node[0] == "range":
            arange_node = arange_node[0]
        node_len = len(arange_node)
        if 1 <= node_len <= 2:
            # arange ::= location
            # arange ::= DIRECTION
            # arange ::= FUNCNAME
            # arange ::= NUMBER
            # arange ::= ADDRESS
            last_node = arange_node[-1]
            if last_node == "location":
                self.preorder(arange_node[-1])
                self.result = ListRange(last_node.location, None)
            elif last_node == "FUNCNAME":
                self.result = ListRange(
                    Location(None, None, False, last_node.value[:-2], offset=0), None
                )
            elif last_node in ("NUMBER", "OFFSET", "ADDRESS"):
                offset = None
                if last_node == "ADDRESS":
                    assert last_node.value[0] in ["*", "@"]
                    is_address = True
                    value = int(last_node.value[1:])
                    offset = value
                else:
                    is_address = False
                    value = last_node.value

                self.result = ListRange(
                    Location(None, value, is_address, None, offset=offset), None
                )
            else:
                assert last_node == "DIRECTION"
                self.result = ListRange(None, last_node.value)
                pass
            self.prune()
        elif node_len == 3:
            # arange ::= COMMA opt_space location
            # arange ::= location opt_space COMMA
            if arange_node[0] == "COMMA":
                assert arange_node[-1] == "location"
                self.preorder(arange_node[-1])
                self.result = ListRange(None, self.result)
                self.prune()
            else:
                assert arange_node[-1] == "COMMA"
                assert arange_node[0] == "location"
                self.preorder(arange_node[0])
                self.result = ListRange(arange_node[0].location, None)
                self.prune()
                pass
        elif node_len == 5:
            # arange ::= location opt_space COMMA opt_space {NUMBER | OFFSET | ADDRESS}
            assert arange_node[2] == "COMMA"
            assert arange_node[-1] in ("NUMBER", "OFFSET", "ADDRESS")
            self.preorder(arange_node[0])
            self.result = ListRange(arange_node[0].location, arange_node[-1].value)
            self.prune()
        else:
            raise RangeError("Something is wrong")
        return

    def n_range(self, range_node):
        node_len = len(range_node)
        if 1 <= node_len <= 2:
            # range ::= location
            # range ::= DIRECTION
            # range ::= FUNCNAME
            # range ::= NUMBER
            # range ::= OFFSET
            last_node = range_node[-1]
            if last_node == "location":
                self.preorder(range_node[-1])
                self.result = ListRange(last_node.location, None)
            elif last_node == "FUNCNAME":
                self.result = ListRange(
                    Location(None, None, False, last_node.value[:-2], offset=None), None
                )
            elif last_node in ("NUMBER", "OFFSET"):
                self.result = ListRange(
                    Location(None, last_node.value, False, None, offset=None), None
                )
            else:
                assert last_node == "DIRECTION"
                self.result = ListRange(None, last_node.value)
                pass
            self.prune()
        elif node_len == 3:
            # range ::= COMMA opt_space location
            # range ::= location opt_space COMMA
            if range_node[0] == "COMMA":
                assert range_node[-1] == "location"
                self.preorder(range_node[-1])
                self.result = ListRange(None, self.result)
                self.prune()
            else:
                assert range_node[-1] == "COMMA"
                assert range_node[0] == "location"
                self.preorder(range_node[0])
                self.result = ListRange(range_node[0].location, None)
                self.prune()
                pass
        elif node_len == 5:
            # range ::= location opt_space COMMA opt_space {NUMBER|OFFSET}
            assert range_node[2] == "COMMA"
            assert range_node[-1] in ("NUMBER", "OFFSET")
            self.preorder(range_node[0])
            self.result = ListRange(range_node[0].location, range_node[-1].value)
            self.prune()
        else:
            raise RangeError("Something is wrong")
        return

    def default(self, node):
        if node not in frozenset(
            (
                """opt_space tokens token bp_start range_start arange_start
                                  IF FILENAME COLON COMMA SPACE DIRECTION""".split()
            )
        ):
            assert False, "Something's wrong: you missed a rule for %s" % node.kind

    def traverse(self, node):
        return self.preorder(node)


def build_bp_expr(string, show_tokens=False, show_ast=False, show_grammar=False):
    parser_debug = {
        "rules": False,
        "transition": False,
        "reduce": show_grammar,
        "errorstack": None,
        "dups": False
        # 'context': True, 'dups': True
    }
    parsed = parse_bp_location(
        string, show_tokens=show_tokens, parser_debug=parser_debug
    )
    assert parsed == "bp_start"
    if show_ast:
        print(parsed)
    walker = LocationGrok(string)
    walker.traverse(parsed)
    bp_expr = walker.result
    if isinstance(bp_expr, Location):
        bp_expr = BPLocation(bp_expr, None)
    location = bp_expr.location
    assert (
        location.line_number is not None
        or location.offset is not None
        or location.method
    )
    return bp_expr


def build_range(string, show_tokens=False, show_ast=False, show_grammar=False):
    parser_debug = {
        "rules": False,
        "transition": False,
        "reduce": show_grammar,
        "errorstack": None,
        "context": False,
        "dups": True,
    }
    parsed = parse_range(string, show_tokens=show_tokens, parser_debug=parser_debug)
    if show_ast:
        print(parsed)
    assert parsed == "range_start"
    walker = LocationGrok(string)
    walker.traverse(parsed)
    list_range = walker.result
    return list_range


def build_location(string, show_tokens=False, show_ast=False, show_grammar=False):
    parser_debug = {
        "rules": False,
        "transition": False,
        "reduce": show_grammar,
        "errorstack": None,
        "context": False,
        "dups": True,
    }
    parsed = parse_location(
        "location", string, show_tokens=show_tokens, parser_debug=parser_debug
    )
    if show_ast:
        print(parsed)
    assert parsed == "location"
    walker = LocationGrok(string)
    walker.traverse(parsed)
    return walker.result


# FIXME: DRY with build_range
def build_arange(string, show_tokens=False, show_ast=False, show_grammar=False):
    parser_debug = {
        "rules": False,
        "transition": False,
        "reduce": show_grammar,
        "errorstack": None,
        "context": True,
        "dups": True,
    }
    parsed = parse_arange(string, show_tokens=show_tokens, parser_debug=parser_debug)
    if show_ast:
        print(parsed)
    assert parsed == "arange_start"
    walker = LocationGrok(string)
    walker.traverse(parsed)
    list_range = walker.result
    return list_range


if __name__ == "__main__":
    # FIXME: make sure the below is in a test
    def doit(fn, line):
        print("=" * 30)
        print(line)
        print("+" * 30)
        try:
            result = fn(line)
            print(result)
        except ScannerError as e:
            print("Scanner error")
            print(e.text)
            print(e.text_cursor)
        except PLocationError as e:
            print("Parser error at or near")
            print(e.text)
            print(e.text_cursor)

    lines = """
    @4
    *4
    /tmp/foo.py:12
    12
    ../foo.py:5
    gcd()
    foo.py:5 if x > 1
    """.splitlines()
    for line in lines:
        if not line.strip():
            continue
        print("=" * 30)
        print(line)
        print("+" * 30)
        bp_expr = build_bp_expr(line)
        print(bp_expr)
    # lines = (
    #     "/tmp/foo.py:12 , 5",
    #     "-"
    #     "+"
    #     "../foo.py:5",
    #     "../foo.py:5 ",
    #     ", 5",
    #     "6 , +2",
    #     ", /foo.py:5",
    #     )
    # lines = (
    #     "*10",
    #     "@10",
    #     "/tmp/foo.py:1, *5",
    #     "../foo.py:0,  +5",
    #     "../foo.py:5,  *30",
    #     "*0, *10",
    #     ", /foo.py:5",
    #     )
    # for line in lines:
    #     line = line.strip()
    #     if not line:
    #         continue
    #     doit(build_arange, line)



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020, 2023-2024
#  Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import os
import sys
from getopt import GetoptError, getopt

from trepan.lib.file import executable
from trepan.misc import wrapped_lines

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class RestartCommand(DebuggerCommand):
    """**restart** [*options*]

    Restart debugger and program via an *exec()* call. All state is lost,
    and new copy of the debugger is used.

    *options* are:

       -p | --python - Add Python interpreter to invocation
       -e | --env    - Use current Environment settings

    Normally invocation is what was stored in sys.args. However
    it can happen that sys.argv[0] can be not marked  executable in the OS.
    When this happens, we need to add "python" at the beginning. Use the "-p" option
    do that using the current Python interpreter running.

    Sometimes in the invocation various environment variables are
    set and these too might get lost in the execv. To pre

    See also:
    ---------

    `run` for another way to restart the debugged program.

    See `quit`, `exit` or `kill` for termination commands."""

    short_help = "(Hard) restart of program via execv()"

    # FIXME: add mechanism for adding python interpreter.
    DebuggerCommand.setup(locals(), category="support", max_args=2)

    def run(self, args):
        # Not sure if we need --env. But for now we'll try it.
        try:
            opts, args = getopt(args[1:], "hep", "help env python".split())
        except GetoptError as err:
            # print help information and exit:
            print(str(err))  # will print something like "option -a not recognized"
            return

        restart_options = {"env": False, "python": False}
        for o, _ in opts:
            if o in ("-h", "--help"):
                self.proc.commands["help"].run(["help", "restart"])
                return
            elif o in ("-e", "--env"):
                restart_options["env"] = True
            elif o in ("-p", "--python"):
                restart_options["python"] = True
            else:
                self.errmsg(f"unhandled option '{o}'")
            pass

        sys_argv = self.debugger.restart_argv()
        if sys_argv and len(sys_argv) > 0:
            program_file = sys_argv[0]
            if not restart_options["python"]:
                if not executable(sys_argv[0]):
                    self.errmsg(f'File "{program_file}" is not marked executable.')
                    return

            exec_suffix = "e" if restart_options["env"] else ""
            confirmed = self.confirm(f"Restart (execvp{exec_suffix})", False)
            if confirmed:
                if restart_options["python"]:
                    sys_argv.insert(0, sys.executable)

                self.msg(
                    wrapped_lines(
                        "Re exec'ing:", repr(sys_argv), self.settings["width"]
                    )
                )
                # Run atexit finalize routines. This seems to be Kosher:
                # http://mail.python.org/pipermail/python-dev/2009-February/085791.html # NOQA
                try:
                    atexit._run_exitfuncs()
                except Exception:
                    pass

                if restart_options["env"]:
                    os.execvpe(program_file, sys_argv, os.environ)
                else:
                    os.execvp(program_file, sys_argv)
                pass
            pass
        else:
            self.errmsg("No executable file and command options recorded.")
            pass
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.mock import dbg_setup

    d, cp = dbg_setup()
    command = RestartCommand(cp)
    command.run([])

    if len(sys.argv) > 1:
        # Strip of arguments so we don't loop in exec.
        d.orig_sys_argv = ["python", sys.argv[0]]
        command.run([])
        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2015, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.lib import complete as Mcomplete


class InfoSignals(Mbase_subcmd.DebuggerSubcommand):
    r"""**info signals** [*signal-name*]

    **info signals** *

    Show information about how debugger treats signals to the program.
    Here are the boolean actions we can take:

     * Stop: enter the debugger when the signal is sent to the debugged program

     * Print: print that the signal was received

     * Stack: show a call stack

     * Pass: pass the signal onto the program

    If *signal-name* is not given, we the above show information for all
    signals. If '*' is given we just give a list of signals.
    """

    min_abbrev = 3  # info sig
    need_stack = False
    short_help = "What debugger does when program gets various signals"

    def complete(self, prefix):
        completions = sorted(["*"] + self.debugger.sigmgr.siglist)
        return Mcomplete.complete_token(completions, prefix)

    def run(self, args):
        if len(args) > 0 and args[0] == "*":
            self.msg(self.columnize_commands(self.debugger.sigmgr.siglist))
        else:
            self.debugger.sigmgr.info_signal(["signal"] + args)
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock, info as Minfo

    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoSignals(i)
    # sub.run([])
    # sub.run(['*'])
    pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2012-2013, 2015, 2017, 2020, 2023 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.eval import extract_expression


class EvalCommand(DebuggerCommand):
    """**eval** *python-statement*

    Run *python-statement* in the context of the current frame.

    If no string is given, we run the string from the current source code
    about to be run. If the command ends `?` (via an alias) and no string is
    given, the following translations occur:

       assert = <expr>       => <expr>
       {if|elif} <expr> :    => <expr>
       while <expr> :        => <expr>
       return <expr>         => <expr>
       for <var> in <expr> : => <expr>
       and <expr>            => <expr>
       or <expr>             => <expr>
       <var> = <expr>        => <expr>

    The above is done via regular expression matching. No fancy parsing is
    done, say, to look to see if *expr* is split across a line or whether
    var an assignment might have multiple variables on the left-hand side.

    Examples:
    ---------

        eval 1+2  # 3
        eval      # Run current source-code line
        eval?     # but strips off leading 'if', 'while', ..
                  # from command

    See also:
    ---------

    `deval`, `set autoeval`, `pr`, `pp` and `examine`."""

    aliases = ("eval?", "?")
    short_help = "Print value of expression EXP"

    DebuggerCommand.setup(locals(), category="data", need_stack=True)

    def run(self, args):
        if 1 == len(args):
            if self.proc.current_source_text:
                text = self.proc.current_source_text.rstrip("\n")
                if "?" == args[0][-1]:
                    text = extract_expression(text)
                    self.msg("eval: %s" % text)
                    pass
            else:
                self.errmsg("Don't have program source text")
                return
        else:
            text = self.proc.current_command[len(self.proc.cmd_name) :]
            pass
        text = text.strip()
        try:
            self.proc.exec_line(text)
        except Exception:
            pass


if __name__ == "__main__":
    import inspect
    from trepan.debugger import Trepan

    d = Trepan()
    cp = d.core.processor
    cp.curframe = inspect.currentframe()
    command = EvalCommand(cp)
    me = 10

    # command.run([command.name, '1+2'])
    # command.run([command.name, 'if 5: x=1'])
    pass



"""Unit test for trepan.inout.tcp*"""
from trepan.inout.tcpclient import TCPClient
from trepan.inout.tcpserver import TCPServer


def test_client_server():
    client = None
    server = None
    try:
        try:
            server = TCPServer(opts={"open": True})
        except Exception:
            print("Skipping because of server open failure")
            return
        print("Server port is %s" % server.PORT)
        try:
            client = TCPClient(opts={"open": True, "PORT": server.PORT})
        except IOError:
            print("Skipping because of client open failure")
            return
        for line in ["one", "two", "three"]:
            server.writeline(line)
            assert line == client.read_msg().rstrip("\n")
            pass

        for line in ["four", "five", "six"]:
            client.writeline(line)
            assert line == server.read_msg().rstrip("\n")
            pass

    finally:
        if client:
            client.close()
        if server:
            server.close()
    return



"""
Helper module for functional testing.
"""
import re
import sys

from trepan.debugger import Trepan
from trepan.inout.stringarray import StringArrayInput, StringArrayOutput


def strarray_setup(debugger_cmds):
    """Common setup to create a debugger with stringio attached"""
    stringin = StringArrayInput(debugger_cmds)
    stringout = StringArrayOutput()
    d_opts = {"input": stringin, "output": stringout}
    d = Trepan(d_opts)
    d.settings["basename"] = True
    d.settings["different"] = False
    d.settings["autoeval"] = False
    d.settings["highlight"] = "plain"
    return d


trepan_prompt = re.compile(r"^.. \d+.*\n\(trepan3k(:.+)?\) ")
trepan_loc = re.compile(r"^\(.+:\d+\): ")


def filter_line_cmd(a: list) -> list:
    """Return output with source lines prompt and command removed"""
    # Remove extra leading spaces.
    # For example:
    # -- 42         y = 5
    # becomes
    # -- y = 5
    a1 = [re.sub(r"^(..) \d+\s+", r"\1 ", s) for s in a if re.match(r"^.. \d+\s+", s)]

    # Remove debugger prompts
    # For example:
    #  (trepan3k)
    a2 = [re.sub(r"\n\(trepan3k\) .*", "", s) for s in a1]

    # Remove locations (test-next.py:41): test_next_between_fn
    a3 = [re.sub(r"\n\(.*:\d+\):.*", "", s) for s in a2]

    # Strip leading and trailing blanks.
    a4 = [s.strip() for s in a3]

    return a4


def get_lineno() -> int:
    """Return the caller's line number"""
    caller = sys._getframe(1)
    return caller.f_lineno


def compare_output(obj, right: list, d):
    got = filter_line_cmd(d.intf[-1].output.output)
    if got != right:
        for i in range(len(got)):
            if i < len(right) and got[i] != right[i]:
                print("! ", got[i])
            else:
                print("  ", got[i])
                pass
            pass
        print("-" * 10)
        for i in range(len(right)):
            if i < len(got) and got[i] != right[i]:
                print("! ", right[i])
            else:
                print("  ", right[i])
                pass
            pass
        pass
    obj.assertEqual(right, got)
    return


# Demo it
if __name__ == "__main__":
    print(get_lineno())
    pass



"""Unit test for trepan.processor.cmdproc"""

import inspect
from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.cmdproc import arg_split, resolve_name

errors = []
msgs = []


def errmsg(msg_str: str):
    errors.append(msg_str)
    return


def msg(msg_str: str):
    msg.append(msg_str)
    return


def set_up():
    d, cp = setup_unit_test_debugger()
    cp.intf[-1].msg = msg
    cp.intf[-1].errmsg = errmsg
    return d, cp


def test_basic():
    _, cp = set_up()
    # We assume there's at least one command
    assert len(cp.commands) > 0
    assert len(cp.aliases) > 0

    # In fact we assume there is a 'quit' command...
    assert "quit" == resolve_name(cp, "quit")
    #   with alias 'q'

    assert "quit" == resolve_name(cp, "q")
    # processor.cmdproc.print_source_line(self.msg, 100,
    #                                    'source_line_test.py')

    assert 3 == cp.eval("1+2")
    return


def test_class_vars():
    """See that each command has required attributes defined.  Possibly in
    a strongly-typed language we would not need to do much of this."""

    _, cp = set_up()

    for cmd in list(cp.commands.values()):
        name = cmd.__class__
        for attr in ["aliases", "min_args", "max_args", "name", "need_stack"]:
            assert hasattr(cmd, attr), f"{name} command should have a {attr} attribute"
            pass

        for attr in ["category", "short_help"]:
            assert hasattr(cmd, attr), f"{name} command should have a {attr} attribute"
            value = getattr(cmd, attr)
            assert isinstance(
                value, str
            ), f"{name} command {attr} attribute should be a string"
            pass

        assert isinstance(cmd.name, str)
        assert isinstance(
            cmd.aliases, tuple
        ), f"{repr(cmd.aliases)} aliases should be a tuple type"
        for value in cmd.aliases:
            assert isinstance(value, str), f"{name} command aliases should be strings"

        if cmd.min_args is not None:
            if cmd.max_args is not None:
                assert cmd.min_args <= cmd.max_args, "%s min_args: %d, max_args: %d" % (
                    name,
                    cmd.min_args,
                    cmd.max_args,
                )
                pass
            pass
        pass


def test_args_split():
    for test, expect in (
        ("Now is the time", [["Now", "is", "the", "time"]]),
        ("Now is the time ;;", [["Now", "is", "the", "time"], []]),
        ("Now is 'the time'", [["Now", "is", "'the time'"]]),
        (
            "Now is the time ;; for all good men",
            [["Now", "is", "the", "time"], ["for", "all", "good", "men"]],
        ),
        (
            "Now is the time ';;' for all good men",
            [["Now", "is", "the", "time", "';;'", "for", "all", "good", "men"]],
        ),
    ):
        assert expect == arg_split(test)
        pass
    return


def test_preloop_hooks():
    _, cp = setup_unit_test_debugger()
    fn = cp.commands["list"]
    assert 0 == len(cp.preloop_hooks), "Should start out with no preloop hooks"
    assert not cp.remove_preloop_hook(
        fn
    ), "Should not be able to return a non-existent hook"
    cp.add_preloop_hook(fn)
    assert 1 == len(cp.preloop_hooks), "Should now have one preloop hook added"
    assert cp.remove_preloop_hook(fn)
    assert 0 == len(cp.preloop_hooks), "Should be back to no preloop hooks"
    # FIXME try adding and running a couple of hooks.
    return


def test_populate_commands():
    """Test that we are creating instances for all of classes of files
    in the command directory ."""
    _, cp = setup_unit_test_debugger()
    for i in cp.cmd_instances:
        if hasattr(i, "aliases"):
            assert isinstance(i.aliases, tuple), f"not tuple {repr(i.aliases)}."

            assert [] == [
                item for item in i.aliases if str != type(item)
            ], f"elements of tuple should be strings {repr(i.aliases)}"
            pass
        pass
    return


def test_get_commands_aliases():
    """Test that the command processor finds a command, alias, and method"""
    _, cp = setup_unit_test_debugger()
    assert "quit" in list(cp.commands.keys())
    assert "quit" == cp.aliases["q"]
    assert inspect.ismethod(cp.commands["quit"].run)
    return


def test_resolve_name():
    """Test that the command processor finds a command, alias, and method"""

    _, cp = setup_unit_test_debugger()
    assert resolve_name(cp, "quit")
    assert resolve_name(cp, "q")
    return



"""Unit test for trepan.interfaces.user"""

from io import StringIO

from trepan.inout.input import DebuggerUserInput
from trepan.interfaces.user import UserInterface


def readline(answer):
    return answer


def test_confirm():
    """Test interface.user.UserInterface.confirm()"""

    inp = StringIO("")
    user_input = DebuggerUserInput(inp)

    def no_history():
        return False

    user_input.use_history = no_history
    u = UserInterface(inp=user_input, opts={"debugger_name": "trepan-pytest"})

    for s in ["y", "Y", "Yes", "  YES  "]:

        def readline_with_prompt(prompt=None):
            return readline(s)

        u.input.readline = readline_with_prompt

        assert u.confirm("Testing", True)
        pass
    for s in ["n", "N", "No", "  NO  "]:

        def readline_with_prompt(prompt=None):
            return readline(s)

        u.input.readline = readline_with_prompt
        assert not u.confirm("Testing", True)
        pass
    # FIXME: Add checking default values. Checking looping
    # values
    return


# FIXME: more thorough testing of other routines in user.



#!/usr/bin/env python3
import sys

from setuptools import find_packages, setup

SYS_VERSION = sys.version_info[0:2]
if not ((3, 1) <= SYS_VERSION < (3, 13)):
    mess = "Python Versions 3.1 to 3.12 are supported only in this package."
    if (2, 4) <= SYS_VERSION <= (2, 7):
        mess += "\nFor your Python, version %s, See trepan2" % sys.version[0:3]
    elif SYS_VERSION < (2, 4):
        mess += "\nFor your Python, version %s, see pydb" % sys.version[0:3]
    print(mess)
    raise Exception(mess)

__import__("pkg_resources")

packages = find_packages()

setup(
    data_files=[
        (
            "trepan/processor/command/help",
            [
                "trepan/processor/command/help/arange.rst",
                "trepan/processor/command/help/command.rst",
                "trepan/processor/command/help/examples.rst",
                "trepan/processor/command/help/filename.rst",
                "trepan/processor/command/help/location.rst",
                "trepan/processor/command/help/range.rst",
                "trepan/processor/command/help/suffixes.rst",
            ],
        )
    ],
    packages=packages,
)



# -*- coding: utf-8 -*-
#   Copyright (C) 2009-2010, 2013, 2015 Rocky Bernstein rocky@gnu.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetSkip(Mbase_subcmd.DebuggerSetBoolSubcommand):
    """**set skip** [ **on** | **off** ]


    Set stopping before *def* or *class* (function or class) statements.

    Classes may have many methods and stand-alone programs may have many
    functions. Often there isn't much value to stopping before defining a
    new function or class into Python's symbol table. (More to the point,
    it can be an annoyance.) However if you do want this, for example
    perhaps you want to debug methods is over-writing one another, then
    set this off

    See also:
    ---------

    `show skip`"""

    in_list = True
    min_abbrev = len("sk")  # Min 'set sk'
    short_help = "Set stopping before def or class statements"
    # FIXME allow individual setting for class and skip.
    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(SetSkip)
    pass



"""Unit test for trepan.processor.command.info_sub.files"""
import test.unit.cmdhelper as cmdhelper
from test.unit.cmdhelper import reset_output, setup_unit_test_debugger

from trepan.processor.command.info import InfoCommand
from trepan.processor.command.info_subcmd import files as MinfoFile

last_was_newline = True


# FIXME: put in a more common place
# Possibly fix up Mock to include this
def setup_io(command):
    reset_output()
    command.msg = cmdhelper.msg
    command.errmsg = cmdhelper.errmsg
    command.msg_nocr = msg_nocr
    return


def msg_nocr(msg_str: str):
    global last_was_newline
    if last_was_newline:
        cmdhelper.msgs.append("")
        pass
    cmdhelper.msgs[-1] += msg_str
    last_was_newline = len(msg_str) == 0
    return


def test_info_files():
    """Test processor.command.info_sub.files.run()"""

    # Set up code.
    _, cp = setup_unit_test_debugger()
    command = InfoCommand(cp, "info")

    sub = MinfoFile.InfoFiles(command)
    setup_io(sub)

    # Run with no options. File should not be cached initially
    reset_output()
    sub.run([])

    assert len(cmdhelper.msgs) == 2
    assert cmdhelper.msgs[0].index("is not cached in debugger") > 0
    assert cmdhelper.msgs[1].startswith("Canonic name:")

    # Run again. File should now be cached.

    reset_output()
    sub.run([])
    assert len(cmdhelper.msgs) == 2
    assert cmdhelper.msgs[0].index("is cached in debugger") > 0
    assert cmdhelper.msgs[1].startswith("Canonic name:")
    assert "width" in sub.settings

    # Test all sub options with different line-widths settings
    # We check from big line width to small and that
    # the number of lines only increases as we get smaller.
    last_lengths = [0, 0, 0]
    for width in (200, 80):
        sub.settings["width"] = width
        i = 0

        # Try running with all non-compound options.
        # Accumulate all messages to compare with option "all"
        all_msgs = []
        first_msg_length = 0

        # Order has to be in the same order as appear for option "all",
        # since  we will be comparing against that.
        for option, expect_str in (
            ("size", "File has "),
            ("sha1", "SHA1 is"),
            ("brkpts", "Possible breakpoint line numbers"),
        ):
            reset_output()
            sub.run([__file__, option])
            assert len(cmdhelper.msgs) >= 3
            assert cmdhelper.msgs[2].index(expect_str) >= 0
            assert len(cmdhelper.msgs) >= last_lengths[i]

            #
            if i == 0:
                all_msgs += cmdhelper.msgs
                first_msg_length = len(cmdhelper.msgs)
            else:
                all_msgs += cmdhelper.msgs[first_msg_length - 1 :]
            i += 1
            pass

        # Run option "all" and compare with previous results
        reset_output()
        sub.run([__file__, "all"])
        assert cmdhelper.msgs == all_msgs

    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from getopt import getopt, GetoptError

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.misc import pretty_modfunc_name
from pyficache import cache_code_lines


class InfoOffsets(DebuggerSubcommand):
    """**info lines** [-n *function-or-module*]

    Show line - function/offset information
    Use -n *function-or-module* to filter results.

    Examples
    --------

        (trepan3k) info lines
        Line - (fn, start offset) table for test/example/gcd.py
          10: <module> @0         21: check_args() @84     36: gcd() @30
          11: <module> @4         22: check_args() @106    37: gcd() @50
          13: <module> @12        23: check_args() @116    38: gcd() @54
          14: check_args() @0     24: check_args() @122    40: <module> @28
          16: check_args() @14    26: <module> @20         41: <module> @36
          17: check_args() @22    30: gcd() @0             43: <module> @42
          18: check_args() @36    31: gcd() @8             44: <module> @60
          19: check_args() @38    34: gcd() @18            45: <module> @84
          20: check_args() @70    35: gcd() @26
        (trepan3k) info lines -n <module>
          10: <module> @0    11: <module> @4   13: <module> @12
          40: <module> @28   26: <module> @20  41: <module> @36
          43: <module> @42   44: <module> @60  45: <module> @84
        (trepan3k) info lines -n gcd
          30: gcd() @0   31: gcd() @8   34: gcd() @18
          35: gcd() @26  36: gcd() @30  37: gcd() @50
          38: gcd() @54

    See also:
    ---------
    `info line`, `info offsets`, `info file`, `info program`, and `info frame`"""

    min_abbrev = 5
    max_args = 2
    need_stack = False
    short_help = "Show line offset information for a file or module"

    def run(self, args):
        """Current line number in source file"""
        # info line <loc>
        try:
            opts, args = getopt(
                args,
                "hn:",
                ["help", "name"],
            )
        except GetoptError:
            # print help information and exit:
            self.errmsg(
                str(sys.exc_info()[1])
            )  # will print something like "option -a not recognized"
            return

        name = None
        for o, a in opts:
            if o in ("-h", "--help"):
                self.proc.commands["help"].run(["help", "info", "lines"])
                return
            elif o in ("-n", "--name"):
                name = a
            else:
                self.errmsg("unhandled option '%s'" % o)
            pass
        pass

        curframe = self.proc.curframe
        if not curframe:
            self.errmsg("No line number information available.")
            return

        # No line number. Use current frame line number
        filename = curframe.f_code.co_filename
        file_info = cache_code_lines(
            filename, toplevel_only=False, include_offsets=True
        )
        if file_info:
            self.section("Line - (fn, start offset) table for %s" % filename)
            lines = []
            for line_number, line_info in file_info.line_numbers.items():
                if not name or any(li.name == name for li in line_info):
                    lines.append(
                        "%4d: %s"
                        % (
                            line_number,
                            ", ".join(
                                [
                                    "%s @%d"
                                    % (pretty_modfunc_name(li.name), li.offsets[0])
                                    for li in line_info
                                    if not name or li.name == name
                                ]
                            ),
                        )
                    )
            m = self.columnize_commands(list(sorted(lines)))
            self.msg(m)
        else:
            self.errmsg("haven't recorded info for offset file %s" % filename)
            pass
        pass

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock
    from trepan.processor.command.info import InfoCommand
    from trepan.debugger import Trepan

    d = Trepan()
    d, cp = mock.dbg_setup(d)
    i = InfoCommand(cp)
    sub = InfoOffsets(i)
    import inspect

    cp.curframe = inspect.currentframe()
    for width in (80, 200):
        sub.settings["width"] = width
        sub.run([])
        pass
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2010, 2013, 2015, 2018, 2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""A base class for a debugger interface."""

import sys

NotImplementedMessage = "This method must be overridden in a subclass"

__all__ = ["TrepanInterface"]


class TrepanInterface:
    """
    A debugger interface handles the communication or interaction with between
    the program and the outside portion which could be
        - a user,
        - a front-end that talks to a user, or
        - another interface in another process or computer
    """

    def __init__(self, inp=None, out=None):
        self.input = inp or sys.stdin
        self.output = out or sys.stdout
        self.interactive = False
        return

    def close(self):
        """Closes all input and/or output"""
        raise NotImplementedError(NotImplementedMessage)
        return

    def confirm(self, prompt, default):
        """Called when a dangerous action is about to be done to make sure
        it's okay. `prompt' is printed; user response is returned."""
        raise NotImplementedError(NotImplementedMessage)

    def errmsg(self, str, prefix="** "):
        """Common routine for reporting debugger error messages."""
        raise NotImplementedError(NotImplementedMessage)

    def finalize(self, last_wishes=None):
        raise NotImplementedError(NotImplementedMessage)

    def msg(self, msg):
        """used to write to a debugger that is connected to this
        server; `str' written will have a newline added to it
        """
        if hasattr(self.output, "writeline"):
            self.output.writeline(msg)
        elif hasattr(self.output, "writelines"):
            self.output.writelines(msg + "\n")
            pass
        return

    def msg_nocr(self, msg):
        """used to write to a debugger that is connected to this
        server; `str' written will not have a newline added to it
        """
        self.output.write(msg)
        return

    def read_command(self, prompt):
        raise NotImplementedError(NotImplementedMessage)

    def readline(self, prompt, add_to_history=True):
        raise NotImplementedError(NotImplementedMessage)

    pass



# Copyright (C) 2013, 2015-2018, 2020-2021, 2023 Rocky Bernstein
# <rocky@gnu.org>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger packaging information"""

# To the extent possible we make this file look more like a
# configuration file rather than code like setup.py. I find putting
# configuration stuff in the middle of a function call in setup.py,
# which for example requires commas in between parameters, is a little
# less elegant than having it here with reduced code, albeit there
# still is some room for improvement.

import os.path as osp
import sys

decompiler = "uncompyle6 >= 3.9.2"

SYS_VERSION = sys.version_info[0:2]
if SYS_VERSION <= (3, 2):
    pygments_version = "== 1.6"
else:
    pygments_version = ">= 2.2.0"
    if (3, 7) <= SYS_VERSION < (3, 9):
        decompiler = "decompyle3 >= 3.8.0"


# Python-version | package  | last-version |
# ------------------------------------------
# 3.2            | pip      | 8.1.2        |
# 3.2            | pygments | 1.6          |
# 3.3            | pip      | 10.0.1       |
# 3.4            | pip      | 19.1.1       |

# Things that change more often go here.
copyright = (
    """Copyright (C) 2013, 2015-2021, 2023-2024 Rocky Bernstein <rocky@gnu.org>."""
)
classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Topic :: Software Development :: Debuggers",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: Implementation :: PyPy",
]

# The rest in alphabetic order
author = "Rocky Bernstein"
author_email = "rocky@gnu.org"

entry_points = {
    "console_scripts": [
        "trepan3k   = trepan.__main__:main",
        "trepan3kc  = trepan.client:main",
    ]
}

ftp_url = None
install_requires = [
    "columnize >= 0.3.10",
    "pyficache >= 2.3.0",
    "xdis >= 6.1.1,<6.2.0",
    "pygments %s" % pygments_version,
    "spark_parser >= 1.8.9, <1.9.0",
    "tracer >= 0.3.2",
    "term-background >= 1.0.1",
]

if decompiler is not None:
    install_requires.append(decompiler)

license = "GPL3"
mailing_list = "python-debugger@googlegroups.com"
modname = "trepan3k"
py_modules = []
short_desc = "GDB-like Python Debugger in the Trepan family"


def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)


def read(*rnames):
    return open(osp.join(get_srcdir(), *rnames)).read()


# version.py sets variable VERSION.
VERSION = None
exec(read("trepan", "version.py"))
web = "http://github.com/rocky/python3-trepan/"

# tracebacks in zip files are funky and not debuggable
zip_safe = False


long_description = read("README.rst") + "\n"



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys

import os.path as osp

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.complete import complete_token
from trepan.lib.file import readable
from trepan.interfaces.script import ScriptInterface


class SourceCommand(DebuggerCommand):
    """**source** [**-v**][**-Y**|**-N**][**-c**] *file*

    Read debugger commands from a file named *file*.  Optional *-v* switch
    (before the filename) causes each command in *file* to be echoed as it
    is executed.  Option *-Y* sets the default value in any confirmation
    command to be "yes" and *-N* sets the default value to "no".

    Note that the command startup file `.trepan3krc` is read automatically
    via a *source* command the debugger is started.

    An error in any command terminates execution of the command file
    unless option `-c` is given."""

    short_help = "Read and run debugger commands from a file"

    DebuggerCommand.setup(locals(), category="support", max_args=1)

    def complete(self, prefix):
        # files = Readline::FILENAME_COMPLETION_PROC.call(prefix) || []
        opts = ["-v", "-Y", "-N", "-c"]  # + files
        return complete_token(opts, prefix)

    def run(self, args):
        parms = args[1:-1]
        opts = {}
        for arg in parms:
            if arg == "-v":
                opts["verbose"] = True
            elif arg == "-Y":
                opts["confirm_val"] = True
            elif arg == "-N":
                opts["confirm_val"] = False
            elif arg == "-c":
                opts["abort_on_error"] = False
            pass
        filename = args[-1]

        expanded_file = osp.expanduser(filename)
        if not readable(expanded_file):
            self.errmsg("Debugger command file '%s' is not a readable file" % filename)
            return False

        # Push a new interface.
        script_intf = ScriptInterface(
            expanded_file, opts=opts, out=self.debugger.intf[-1].output
        )
        self.debugger.intf.append(script_intf)
        return False


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    d, cp = Mmock.dbg_setup()
    command = SourceCommand(cp)
    if len(sys.argv) > 1:
        command.run([sys.argv[1]])
    pass



#!/usr/bin/env python3
"""Unit test for the debugger lib breakpoint"""

import inspect
import os
import re

from trepan.lib.breakpoint import BreakpointManager, checkfuncname


def test_breakpoint():
    """Test breakpoint"""
    bpmgr = BreakpointManager()
    assert 0 == bpmgr.last()
    bp = bpmgr.add_breakpoint("foo", 10, 5)

    assert re.search(r"1\s+breakpoint\s+keep\s+yes .* at .*foo:10", str(bp)), str(bp)
    assert "B" == bp.icon_char()
    assert bp.enabled
    bp.disable()
    assert re.search(r"1\s+breakpoint\s+keep\s+no .* at .*foo:10", str(bp)), str(bp)
    assert not bp.enabled
    assert "b" == bp.icon_char()
    assert 1 == bpmgr.last()
    assert (
        False,
        "Breakpoint number 10 out of range 1..1.",
    ) == bpmgr.delete_breakpoint_by_number(10)
    assert ["1"] == bpmgr.bpnumbers(), "Extracting disabled breakpoint-numbers"
    assert (True, "") == bpmgr.delete_breakpoint_by_number(1)
    assert (False, "Breakpoint 1 previously deleted.") == (
        False,
        "Breakpoint 1 previously deleted.",
    )
    bp2 = bpmgr.add_breakpoint("foo", 5, 10, temporary=True)
    assert "t" == bp2.icon_char()
    assert ["2"] == bpmgr.bpnumbers(), "Extracting breakpoint-numbers"

    count = 3
    for i in range(count):
        bp = bpmgr.add_breakpoint("bar", 10)
    filename = bp.filename
    assert count == len(
        bpmgr.delete_breakpoints_by_lineno(os.path.realpath(filename), 10)
    ), "delete_breakpoints_by_line when there are none"
    assert 0 != len(bpmgr.bplist), "There should still be some breakpoints before reset"
    bpmgr.reset()
    assert 0 == len(bpmgr.bplist), "reset should remove all breakpoints"
    return


def test_checkfuncname():
    """Test breakpoint.checkfuncname()"""

    bpmgr = BreakpointManager()
    frame = inspect.currentframe()
    bp = bpmgr.add_breakpoint("test_funcname", frame.f_lineno + 1, -1)
    assert checkfuncname(bp, frame)

    def foo(brkpt, bpmgr):
        current_frame = inspect.currentframe()
        assert checkfuncname(brkpt, current_frame)
        # current_frame.f_lineno is constantly updated. So adjust for line
        # the difference between the add_breakpoint and the check.
        bp3 = bpmgr.add_breakpoint(
            os.path.realpath(__file__), current_frame.f_lineno + 2, -1, False, None
        )
        assert checkfuncname(bp3, current_frame), str(bp3)
        assert not checkfuncname(bp3, current_frame)
        return

    bp2 = bpmgr.add_breakpoint(None, None, -1, False, None, "foo")
    foo(bp2, bpmgr)
    return



"""Unit test for trepan.processor.command.step"""


from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.command.step import StepCommand

"""Tests StepCommand class"""


def test_step():
    """Test processor.command.step.StepCommand.run()"""
    _, cp = setup_unit_test_debugger()
    command = StepCommand(cp)
    result = command.run(["step", "wrong", "number", "of", "args"])
    assert result is False

    result = command.run(["step", "5"])
    assert result is True
    assert 4 == cp.debugger.core.step_ignore

    result = command.run(["step"])
    assert result is True
    assert 0 == cp.debugger.core.step_ignore

    result = command.run(["step", "1+(2*3)"])
    assert result is True
    assert 6 == cp.debugger.core.step_ignore
    return
